#
# Copyright 2008-2009 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Contains base placeable classes with names based on XLIFF placeables. See the
XLIFF standard for more information about what the names mean.
"""

from translate.storage.placeables.interfaces import (
    Delimiter,
    MaskingPlaceable,
    PairedDelimiter,
    ReplacementPlaceable,
    SubflowPlaceable,
)
from translate.storage.placeables.strelem import StringElem

__all__ = ("Bpt", "Bx", "Ept", "Ex", "G", "It", "Ph", "Sub", "X", "to_base_placeables")


# Basic placeable types.
class Bpt(MaskingPlaceable, PairedDelimiter):
    has_content = True


class Ept(MaskingPlaceable, PairedDelimiter):
    has_content = True


class Ph(MaskingPlaceable):
    has_content = True
    istranslatable = False


class It(MaskingPlaceable, Delimiter):
    has_content = True


class G(ReplacementPlaceable):
    has_content = True


class Bx(ReplacementPlaceable, PairedDelimiter):
    has_content = False
    istranslatable = False

    def __init__(self, id=None, xid=None, **kwargs) -> None:
        # kwargs is ignored
        ReplacementPlaceable.__init__(self, id=id, xid=xid, **kwargs)


class Ex(ReplacementPlaceable, PairedDelimiter):
    has_content = False
    istranslatable = False

    def __init__(self, id=None, xid=None, **kwargs) -> None:
        # kwargs is ignored
        ReplacementPlaceable.__init__(self, id=id, xid=xid, **kwargs)


class X(ReplacementPlaceable, Delimiter):
    has_content = False
    iseditable = False
    isfragile = True
    istranslatable = False

    def __init__(self, id=None, xid=None, **kwargs) -> None:
        ReplacementPlaceable.__init__(self, id=id, xid=xid, **kwargs)


class Sub(SubflowPlaceable):
    has_content = True


def to_base_placeables(tree):
    if not isinstance(tree, StringElem):
        return tree

    base_class = [
        klass
        for klass in tree.__class__.__bases__
        if klass in {Bpt, Ept, Ph, It, G, Bx, Ex, X, Sub}
    ]

    base_class = tree.__class__ if not base_class else base_class[0]

    newtree = base_class()
    newtree.id = tree.id
    newtree.rid = tree.rid
    newtree.xid = tree.xid
    newtree.sub = []

    for subtree in tree.sub:
        newtree.sub.append(to_base_placeables(subtree))

    return newtree



#
# Copyright 2008, 2010 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#

"""
Module for handling Qt Linguist Phrase Book (.qph) files.

Extract from the `Qt Linguist Manual: Translators
<http://doc.trolltech.com/4.3/linguist-translators.html>`_:
.qph Qt Phrase Book Files are human-readable XML files containing standard
phrases and their translations. These files are created and updated by Qt
Linguist and may be used by any number of projects and applications.

A DTD to define the format does not seem to exist, but the following `code
<http://qt.gitorious.org/qt/qt/blobs/4.7/tools/linguist/shared/qph.cpp>`_
provides the reference implementation for the Qt Linguist product.
"""

from lxml import etree

from translate.lang import data
from translate.misc.xml_helpers import safely_set_text
from translate.storage import lisa


class QphUnit(lisa.LISAunit):
    """A single term in the qph file."""

    rootNode = "phrase"
    languageNode = "source"
    textNode = ""
    namespace = ""

    def createlanguageNode(self, lang, text, purpose):  # ty:ignore[invalid-method-override]
        """Returns an xml Element setup with given parameters."""
        assert purpose
        langset = etree.Element(self.namespaced(purpose))
        safely_set_text(langset, text)
        return langset

    def _getsourcenode(self):
        return self.xmlelement.find(self.namespaced(self.languageNode))

    def _gettargetnode(self):
        return self.xmlelement.find(self.namespaced("target"))

    def getlanguageNodes(self):
        """We override this to get source and target nodes."""
        return [
            n for n in [self._getsourcenode(), self._gettargetnode()] if n is not None
        ]

    def addnote(self, text, origin=None, position="append") -> None:
        """Add a note specifically in a "definition" tag."""
        current_notes = self.getnotes(origin)
        self.removenotes(origin)
        note = etree.SubElement(self.xmlelement, self.namespaced("definition"))
        safely_set_text(note, "\n".join(filter(None, [current_notes, text.strip()])))

    def getnotes(self, origin=None):
        # TODO: consider only responding when origin has certain values
        notenode = self.xmlelement.find(self.namespaced("definition"))
        comment = ""
        if notenode is not None:
            comment = notenode.text
        return comment

    def removenotes(self, origin=None) -> None:
        """Remove all the translator notes."""
        note = self.xmlelement.find(self.namespaced("definition"))
        if note is not None:
            self.xmlelement.remove(note)


class QphFile(lisa.LISAfile):
    """Class representing a QPH file store."""

    UnitClass = QphUnit
    Name = "Qt Phrase Book"
    Mimetypes = ["application/x-qph"]
    Extensions = ["qph"]
    rootNode = "QPH"
    bodyNode = "QPH"
    XMLskeleton = """<!DOCTYPE QPH>
<QPH>
</QPH>
"""
    namespace = ""

    def initbody(self) -> None:
        """
        Initialises self.body so it never needs to be retrieved from the XML
        again.
        """
        self.namespace = self.document.getroot().nsmap.get(None, None)
        self.header = self.document.getroot()
        self.body = self.document.getroot()  # The root node contains the units

    def getsourcelanguage(self) -> str:
        """
        Get the source language for this .qph file.

        We don't implement setsourcelanguage as users really shouldn't be
        altering the source language in .qph files, it should be set correctly
        by the extraction tools.

        :return: ISO code e.g. af, fr, pt_BR
        """
        lang = data.normalize_code(self.header.get("sourcelanguage", "en"))
        if lang == "en-us":
            return "en"
        return lang

    def gettargetlanguage(self) -> str:
        """
        Get the target language for this .qph file.

        :return: ISO code e.g. af, fr, pt_BR
        """
        return data.normalize_code(self.header.get("language"))

    def settargetlanguage(self, targetlanguage: str) -> None:
        """
        Set the target language for this .qph file to *targetlanguage*.

        :param targetlanguage: ISO code e.g. af, fr, pt_BR
        """
        if targetlanguage:
            self.header.set("language", targetlanguage)

    def serialize(self, out) -> None:
        """
        Write the XML document to the file `out`.

        We have to override this to ensure mimic the Qt convention:
            - no XML declaration
        """
        self.document.write(
            out, pretty_print=True, xml_declaration=False, encoding="utf-8"
        )



#
# Copyright 2023 Zuza Software Foundation & Anders Kaplan
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#

"""
Convert Markdown files to Gettext PO localization files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/md2po.html
for examples and usage instructions.
"""

from translate.convert import convert
from translate.storage import markdown, po


class MD2POOptionParser(convert.ConvertOptionParser, convert.DocpathMerger):
    def __init__(self) -> None:
        formats = {
            "md": ("po", self._extract_translation_units),
            "markdown": ("po", self._extract_translation_units),
            "txt": ("po", self._extract_translation_units),
            "text": ("po", self._extract_translation_units),
            None: ("po", self._extract_translation_units),
        }
        super().__init__(formats, usetemplates=True, usepots=True, description=__doc__)
        self.add_duplicates_option()
        self.add_multifile_option()

    def _extract_translation_units(
        self,
        inputfile,
        outputfile,
        templatefile,
        duplicatestyle,
        multifilestyle,
    ) -> int:
        if hasattr(self, "outputstore"):
            if templatefile is None:
                self._parse_and_extract(inputfile, self.outputstore)
            else:
                self._merge_with_template(inputfile, templatefile, self.outputstore)
        else:
            store = po.pofile()
            if templatefile is None:
                self._parse_and_extract(inputfile, store)
            else:
                self._merge_with_template(inputfile, templatefile, store)
            store.removeduplicates(duplicatestyle)
            store.serialize(outputfile)
        return 1

    @staticmethod
    def _parse_and_extract(inputfile, outputstore) -> None:
        """Extract translation units from a markdown file and add them to an existing message store (pofile object) without any further processing."""
        parser = markdown.MarkdownFile(inputfile=inputfile)
        for tu in parser.units:
            if not tu.isheader():
                storeunit = outputstore.addsourceunit(tu.source)
                storeunit.addlocations(tu.getlocations())

    def _merge_with_template(self, inputfile, templatefile, outputstore) -> None:
        """Merge translation from inputfile with source from templatefile using docpath matching."""
        self.merge_stores_by_docpath(
            inputfile,
            templatefile,
            outputstore,
            markdown.MarkdownFile,
            filter_header=True,
        )

    def recursiveprocess(self, options) -> None:
        """Recurse through directories and process files. (override)."""
        if options.multifilestyle == "onefile":
            self.outputstore = po.pofile()
            super().recursiveprocess(options)
            if not self.outputstore.isempty():
                outputfile = super().openoutputfile(options, options.output)
                self.outputstore.removeduplicates(options.duplicatestyle)
                self.outputstore.serialize(outputfile)
                if options.output:
                    outputfile.close()
        else:
            super().recursiveprocess(options)

    def isrecursive(self, fileoption, filepurpose="input"):
        """Check if fileoption is a recursive file. (override)."""
        if hasattr(self, "outputstore") and filepurpose == "output":
            return True
        return super().isrecursive(fileoption, filepurpose=filepurpose)

    def checkoutputsubdir(self, options, subdir) -> None:
        """
        Check if subdir under options.output needs to be created, and
        create if necessary. Do nothing if in single-output-file mode. (override).
        """
        if hasattr(self, "outputstore"):
            return
        super().checkoutputsubdir(options, subdir)

    def openoutputfile(self, options, fulloutputpath):
        """Open the output file, or do nothing if in single-output-file mode. (override)."""
        if hasattr(self, "outputstore"):
            return None
        return super().openoutputfile(options, fulloutputpath)


def main(argv=None) -> None:
    parser = MD2POOptionParser()
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2004-2007 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.
#

"""
module that provides modified DOM functionality for our needs.

Note that users of ourdom should ensure that no code might still use classes
directly from minidom, like minidom.Element, minidom.Document or methods such
as minidom.parseString, since the functionality provided here will not be in
those objects.
"""

from xml.dom import (
    EMPTY_NAMESPACE,
    EMPTY_PREFIX,
    XMLNS_NAMESPACE,
    expatbuilder,
    minidom,
)

# helper functions we use to do xml the way we want, used by modified
# classes below


def writexml_helper(self, writer, indent="", addindent="", newl="") -> None:
    """
    A replacement for writexml that formats it like typical XML files.
    Nodes are indented but text nodes, where whitespace can be significant,
    are not indented.
    """
    # indent = current indentation
    # addindent = indentation to add to higher levels
    # newl = newline string
    writer.write(f"{indent}<{self.tagName}")

    attrs = self._get_attributes()
    a_names = sorted(attrs.keys())

    for a_name in a_names:
        writer.write(f' {a_name}="')
        minidom._write_data(writer, attrs[a_name].value)  # ty:ignore[unresolved-attribute]
        writer.write('"')
    if self.childNodes:
        # We need to write text nodes without newline and indentation, so
        # we handle them differently. Note that we here assume that "empty"
        # text nodes can be done away with (see the strip()). Note also that
        # nested tags in a text node (like ph tags in xliff) should also not
        # have newlines and indentation or an extra newline, since that will
        # alter the text node.
        haveText = False
        for childNode in self.childNodes:
            if childNode.nodeType == self.TEXT_NODE and childNode.data.strip():
                haveText = True
                break
        if haveText:
            writer.write(">")
            for node in self.childNodes:
                node.writexml(writer, "", "", "")
            writer.write(f"</{self.tagName}>{newl}")
        else:
            # This is the normal case that we do with pretty layout
            writer.write(f">{newl}")
            for node in self.childNodes:
                if node.nodeType != self.TEXT_NODE:
                    node.writexml(writer, (indent + addindent), addindent, newl)
            writer.write(f"{indent}</{self.tagName}>{newl}")
    else:
        writer.write(f"/>{newl}")


def getElementsByTagName_helper(parent, name):
    """
    A reimplementation of getElementsByTagName as an iterator.

    Note that this is not compatible with getElementsByTagName that returns a
    list, therefore, the class below exposes this through
    yieldElementsByTagName
    """
    for node in parent.childNodes:
        if node.nodeType == minidom.Node.ELEMENT_NODE and (name in {"*", node.tagName}):
            yield node
        if node.hasChildNodes():
            yield from node.getElementsByTagName(name)


def searchElementsByTagName_helper(parent, name, onlysearch):
    """Limits the search to within tags occurring in onlysearch."""
    for node in parent.childNodes:
        if node.nodeType == minidom.Node.ELEMENT_NODE and (name in {"*", node.tagName}):
            yield node
        if node.nodeType == minidom.Node.ELEMENT_NODE and node.tagName in onlysearch:
            yield from node.searchElementsByTagName(name, onlysearch)


def getFirstElementByTagName(node, name):
    results = node.yieldElementsByTagName(name)
    return next(results, None)


def getnodetext(node):
    """Returns the node's text by iterating through the child nodes."""
    if node is None:
        return ""
    return "".join(t.data for t in node.childNodes if t.nodeType == t.TEXT_NODE)


# various modifications to minidom classes to add functionality we like


class DOMImplementation(minidom.DOMImplementation):
    def _create_document(self):
        return Document()


class Element(minidom.Element):
    def yieldElementsByTagName(self, name):
        return getElementsByTagName_helper(self, name)

    def searchElementsByTagName(self, name, onlysearch):
        return searchElementsByTagName_helper(self, name, onlysearch)

    def writexml(self, writer, indent="", addindent="", newl=""):
        return writexml_helper(self, writer, indent, addindent, newl)


class Document(minidom.Document):
    implementation = DOMImplementation()

    def yieldElementsByTagName(self, name):
        return getElementsByTagName_helper(self, name)

    def searchElementsByTagName(self, name, onlysearch):
        return searchElementsByTagName_helper(self, name, onlysearch)

    def createElement(self, tagName):
        e = Element(tagName)
        e.ownerDocument = self
        return e

    def createElementNS(self, namespaceURI, qualifiedName):
        prefix, _localName = minidom._nssplit(qualifiedName)  # ty:ignore[unresolved-attribute]
        e = Element(qualifiedName, namespaceURI, prefix)
        e.ownerDocument = self
        return e


theDOMImplementation = DOMImplementation()

# an ExpatBuilder that allows us to use the above modifications


class ExpatBuilderNS(expatbuilder.ExpatBuilderNS):
    def reset(self) -> None:
        """Free all data structures used during DOM construction."""
        self.document = theDOMImplementation.createDocument(EMPTY_NAMESPACE, None, None)
        self.curNode = self.document
        self._elem_info = self.document._elem_info  # ty:ignore[unresolved-attribute]
        self._cdata = False
        self._initNamespaces()  # ty:ignore[unresolved-attribute]

    def start_element_handler(self, name, attributes) -> None:
        # All we want to do is construct our own Element instead of
        # minidom.Element, unfortunately the only way to do this is to
        # copy this whole function from expatbuilder.py
        if " " in name:
            uri, localname, prefix, qname = expatbuilder._parse_ns_name(self, name)  # ty:ignore[unresolved-attribute]
        else:
            uri = EMPTY_NAMESPACE
            qname = name
            localname = None
            prefix = EMPTY_PREFIX
        node = Element(qname, uri, prefix, localname)
        node.ownerDocument = self.document
        expatbuilder._append_child(self.curNode, node)  # ty:ignore[unresolved-attribute]
        self.curNode = node

        if self._ns_ordered_prefixes:  # ty:ignore[unresolved-attribute]
            for prefix, uri in self._ns_ordered_prefixes:  # ty:ignore[unresolved-attribute]
                if prefix:
                    a = minidom.Attr(
                        expatbuilder._intern(self, f"xmlns:{prefix}"),  # ty:ignore[unresolved-attribute]
                        XMLNS_NAMESPACE,
                        prefix,
                        "xmlns",
                    )
                else:
                    a = minidom.Attr("xmlns", XMLNS_NAMESPACE, "xmlns", EMPTY_PREFIX)
                a.value = uri
                a.ownerDocument = self.document
                expatbuilder._set_attribute_node(node, a)  # ty:ignore[unresolved-attribute]
            del self._ns_ordered_prefixes[:]  # ty:ignore[unresolved-attribute]

        if attributes:
            node._ensure_attributes()  # ty:ignore[unresolved-attribute]
            attrs = node._attrs  # ty:ignore[unresolved-attribute]
            attrsNS = node._attrsNS  # ty:ignore[unresolved-attribute]
            for i in range(0, len(attributes), 2):
                aname = attributes[i]
                value = attributes[i + 1]
                if " " in aname:
                    uri, localname, prefix, qname = expatbuilder._parse_ns_name(  # ty:ignore[unresolved-attribute]
                        self, aname
                    )
                    a = minidom.Attr(qname, uri, localname, prefix)
                    attrs[qname] = a
                    attrsNS[uri, localname] = a
                else:
                    a = minidom.Attr(aname, EMPTY_NAMESPACE, aname, EMPTY_PREFIX)
                    attrs[aname] = a
                    attrsNS[EMPTY_NAMESPACE, aname] = a
                a.ownerDocument = self.document
                a.value = value
                a.ownerElement = node

    if __debug__:
        # This only adds some asserts to the original
        # end_element_handler(), so we only define this when -O is not
        # used.  If changing one, be sure to check the other to see if
        # it needs to be changed as well.

        def end_element_handler(self, name) -> None:
            curNode = self.curNode
            if " " in name:
                uri, localname, prefix, _qname = expatbuilder._parse_ns_name(self, name)  # ty:ignore[unresolved-attribute]
                assert curNode.namespaceURI == uri, (  # ty:ignore[unresolved-attribute]
                    "element stack messed up! (namespace)"
                )
                assert curNode.localName == localname  # ty:ignore[unresolved-attribute]
                assert curNode.prefix == prefix  # ty:ignore[unresolved-attribute]
            else:
                assert curNode.nodeName == name, (  # ty:ignore[unresolved-attribute]
                    "element stack messed up - bad nodeName"
                )
                assert curNode.namespaceURI == EMPTY_NAMESPACE, (  # ty:ignore[unresolved-attribute]
                    "element stack messed up - bad namespaceURI"
                )
            self.curNode = curNode.parentNode  # ty:ignore[invalid-assignment, unresolved-attribute]
            self._finish_end_element(curNode)  # ty:ignore[unresolved-attribute]


# parser methods that use our modified xml classes


def parse(file):
    """Parse a file into a DOM by filename or file object."""
    builder = ExpatBuilderNS()
    if isinstance(file, str):
        with open(file, "rb") as fp:
            result = builder.parseFile(fp)
    else:
        result = builder.parseFile(file)
    return result


def parseString(string):
    """Parse a file into a DOM from a string."""
    builder = ExpatBuilderNS()
    return builder.parseString(string)



#
# Copyright 2002-2011 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""class that handles all header functions for a header in a po file."""

import re
import time

from translate import __version__
from translate.misc.dictutils import cidict

author_re = re.compile(r".*<\S+@\S+>.*\d{4,4}")
nplural_re = re.compile(r"nplurals=(.+?);")
plural_re = re.compile(r"plural=(.+?);?$")

default_header = {
    "Project-Id-Version": "PACKAGE VERSION",
    "PO-Revision-Date": "YEAR-MO-DA HO:MI+ZONE",
    "Last-Translator": "FULL NAME <EMAIL@ADDRESS>",
    "Language-Team": "LANGUAGE <LL@li.org>",
    "Plural-Forms": "nplurals=INTEGER; plural=EXPRESSION;",
}


def parseheaderstring(input):
    """
    Parses an input string with the definition of a PO header and returns
    the interpreted values as a dictionary.
    """
    headervalues = {}
    for line in input.split("\n"):
        if not line or ":" not in line:
            continue
        key, value = line.split(":", 1)
        key = key.strip()
        headervalues[key] = value.strip()
    return headervalues


def tzstring() -> str:
    """Returns the timezone as a string in the format [+-]0000, eg +0200."""
    tzoffset = time.altzone if time.daylight else time.timezone

    hours, minutes = time.gmtime(abs(tzoffset))[3:5]
    if tzoffset > 0:
        hours *= -1
    return f"{hours:+03d}{minutes:02d}"


def format_key(key: str) -> str:
    key = key.replace("_", "-")
    if key.islower():
        return key.title()
    return key


def update(existing: dict[str, str], add: bool = False, **kwargs) -> dict[str, str]:
    """
    Update an existing header dictionary with the values in kwargs, adding
    new values only if add is true.

    :return: Updated dictionary of header entries
    """
    headerargs = {}
    fixedargs = cidict((format_key(key), value) for key, value in kwargs.items())
    removed = set()
    for key in poheader.header_order:
        if key in existing:
            if key in fixedargs:
                headerargs[key] = fixedargs.pop(key)
            else:
                headerargs[key] = existing[key]
            removed.add(key)
        elif add and key in fixedargs:
            headerargs[key] = fixedargs.pop(key)
    headerargs.update(
        (key, value) for key, value in existing.items() if key not in removed
    )
    if add:
        headerargs.update(fixedargs)
    return headerargs


class poheader:
    """
    This class implements functionality for manipulation of po file headers.
    This class is a mix-in class and useless on its own. It must be used from
    all classes which represent a po file.
    """

    x_generator = f"Translate Toolkit {__version__.sver}"

    header_order = [
        "Project-Id-Version",
        "Report-Msgid-Bugs-To",
        "POT-Creation-Date",
        "PO-Revision-Date",
        "Last-Translator",
        "Language-Team",
        "Language",
        "MIME-Version",
        "Content-Type",
        "Content-Transfer-Encoding",
        "Plural-Forms",
        "X-Accelerator-Marker",
        "X-Generator",
        "X-Merge-On",
    ]

    def init_headers(self, charset="UTF-8", encoding="8bit", **kwargs):
        """Sets default values for po headers."""
        # FIXME: we need to allow at least setting target language, pluralforms and generator
        headerdict = self.makeheaderdict(charset=charset, encoding=encoding, **kwargs)
        self.updateheader(add=True, **headerdict)
        return self.header()

    def makeheaderdict(
        self,
        charset="CHARSET",
        encoding="ENCODING",
        project_id_version=None,
        pot_creation_date=None,
        po_revision_date=None,
        last_translator=None,
        language_team=None,
        mime_version=None,
        plural_forms=None,
        report_msgid_bugs_to=None,
        **kwargs,
    ) -> dict[str, str]:
        """
        Create a header dictionary with useful defaults.

        pot_creation_date can be None (current date) or a value (datetime or string)
        po_revision_date can be None (form), False (=pot_creation_date), True (=now),
        or a value (datetime or string)

        :return: Dictionary with the header items
        """

        def format_date(key, value, fallback=None) -> None:
            if value is False and (fallback is not None):
                defaultargs[key] = fallback
            elif value is True or (value is None and fallback is None):
                defaultargs[key] = time.strftime("%Y-%m-%d %H:%M") + tzstring()
            elif isinstance(value, time.struct_time):
                defaultargs[key] = time.strftime("%Y-%m-%d %H:%M", value) + tzstring()
            elif value is not None:
                defaultargs[key] = value

        defaultargs = default_header.copy()

        if project_id_version is not None:
            defaultargs["Project-Id-Version"] = project_id_version

        format_date("POT-Creation-Date", pot_creation_date)
        format_date(
            "PO-Revision-Date",
            po_revision_date,
            defaultargs["POT-Creation-Date"],
        )

        if last_translator is not None:
            defaultargs["Last-Translator"] = last_translator

        if language_team is not None:
            defaultargs["Language-Team"] = language_team

        defaultargs["MIME-Version"] = mime_version or "1.0"

        defaultargs["Report-Msgid-Bugs-To"] = report_msgid_bugs_to or ""

        defaultargs["Content-Type"] = f"text/plain; charset={charset}"
        defaultargs["Content-Transfer-Encoding"] = encoding
        if plural_forms:
            defaultargs["Plural-Forms"] = plural_forms
        else:
            del defaultargs["Plural-Forms"]
        defaultargs["X-Generator"] = self.x_generator

        return update(defaultargs, add=True, **kwargs)

    def header(self):
        """
        Returns the header element, or None. Only the first element is
        allowed to be a header. Note that this could still return an empty
        header element, if present.
        """
        if len(self.units) == 0:  # ty:ignore[unresolved-attribute]
            return None
        candidate = self.units[0]  # ty:ignore[unresolved-attribute]
        if candidate.isheader():
            return candidate
        return None

    def parseheader(self):
        """
        Parses the PO header and returns the interpreted values as a
        dictionary.
        """
        header = self.header()
        if not header:
            return {}
        return parseheaderstring(header.target)

    def updateheader(self, add=False, **kwargs):
        """
        Updates the fields in the PO style header.

        This will create a header if add == True.
        """
        header = self.header()
        if not header:
            if add:
                header = self.makeheader(**kwargs)
                self._insert_header(header)
        else:
            headeritems = update(self.parseheader(), add, **kwargs)
            keys = headeritems.keys()
            if (
                "Content-Type" not in keys
                or "charset=CHARSET" in headeritems["Content-Type"]
            ):
                headeritems["Content-Type"] = "text/plain; charset=UTF-8"
            if (
                "Content-Transfer-Encoding" not in keys
                or "ENCODING" in headeritems["Content-Transfer-Encoding"]
            ):
                headeritems["Content-Transfer-Encoding"] = "8bit"
            headerString = ""
            for key, value in headeritems.items():
                if value is not None:
                    headerString += f"{key}: {value}\n"
            header.target = headerString
            header.markfuzzy(False)  # TODO: check why we do this?
        return header

    def _insert_header(self, header) -> None:
        # we should be using .addunit() or some equivalent in case the
        # unit needs to refer back to the store, etc. This might be
        # subtly broken for POXLIFF, since we don't duplicate the code
        # from lisa::addunit().
        header._store = self
        self.units.insert(0, header)  # ty:ignore[unresolved-attribute]

    def getheaderplural(self):
        """Returns the nplural and plural values from the header."""
        header = self.parseheader()
        pluralformvalue = header.get("Plural-Forms")
        if pluralformvalue is None:
            return None, None
        nplural = nplural_re.findall(pluralformvalue)
        plural = plural_re.findall(pluralformvalue)
        nplural = None if not nplural or nplural[0] == "INTEGER" else nplural[0]
        plural = None if not plural or plural[0] == "EXPRESSION" else plural[0]
        return nplural, plural

    def updateheaderplural(self, nplurals, plural) -> None:
        """Update the Plural-Form PO header."""
        if isinstance(nplurals, str):
            nplurals = int(nplurals)
        self.updateheader(
            add=True, Plural_Forms=f"nplurals={nplurals}; plural={plural};"
        )

    def gettargetlanguage(self):
        """
        Return the target language based on information in the header.

        The target language is determined in the following sequence:
          1. Use the 'Language' entry in the header.
          2. Poedit's custom headers.
          3. Analysing the 'Language-Team' entry.
        """
        header = self.parseheader()
        if lang := header.get("Language"):
            # pylint: disable-next=import-outside-toplevel
            from translate.lang.data import langcode_ire  # noqa: PLC0415

            if langcode_ire.match(lang):
                return lang
        if "X-Poedit-Language" in header:
            # pylint: disable-next=import-outside-toplevel
            from translate.lang import poedit  # noqa: PLC0415

            language = header.get("X-Poedit-Language")
            country = header.get("X-Poedit-Country")
            return poedit.isocode(language, country)
        if "Language-Code" in header:  # Used in Plone files
            return header.get("Language-Code")
        if "Language-Team" in header:
            # pylint: disable-next=import-outside-toplevel
            from translate.lang.team import guess_language  # noqa: PLC0415

            return guess_language(header.get("Language-Team"))
        return None

    def settargetlanguage(self, lang: str) -> None:
        """
        Set the target language in the header.

        This removes any custom Poedit headers if they exist.

        :param lang: the new target language code
        """
        if isinstance(lang, str) and len(lang) > 1:
            self.updateheader(
                add=True, Language=lang, X_Poedit_Language=None, X_Poedit_Country=None
            )

    def getprojectstyle(self) -> str | None:
        """
        Return the project based on information in the header.

        The project is determined in the following sequence:
          1. Use the 'X-Project-Style' entry in the header.
          2. Use 'Report-Msgid-Bug-To' entry
          3. Use the 'X-Accelerator' entry
          4. Use the Project ID
          5. Analyse the file itself (not yet implemented)
        """
        header = self.parseheader()
        project = header.get("X-Project-Style")
        if project is not None:
            return project
        bug_address = header.get("Report-Msgid-Bugs-To")
        if bug_address is not None:
            if "bugzilla.gnome.org" in bug_address:
                return "gnome"
            if "bugs.kde.org" in bug_address:
                return "kde"
        accelerator = header.get("X-Accelerator-Marker")
        if accelerator is not None:
            if accelerator == "~":
                return "openoffice"
            if accelerator == "&":
                return "mozilla"
        project_id = header.get("Project-Id-Version")
        if project_id is not None and "gnome" in project_id.lower():
            return "gnome"
        # TODO Call some project guessing code and probably move all of the above there also
        return None

    def setprojectstyle(self, project_style: str) -> None:
        """
        Set the project in the header.

        :param project_style: the new project
        """
        self.updateheader(add=True, X_Project_Style=project_style)

    def mergeheaders(self, otherstore) -> None:
        """
        Merges another header with this header.

        This header is assumed to be the template.

        :param otherstore: The other store to merge headers from.
        """
        newvalues = otherstore.parseheader()
        retain_list = (
            "Project-Id-Version",
            "PO-Revision-Date",
            "Last-Translator",
            "Language-Team",
            "Plural-Forms",
        )
        retain = {
            key: newvalues[key]
            for key in retain_list
            if newvalues.get(key, None) and newvalues[key] != default_header.get(key)
        }
        self.updateheader(**retain)

    def updatecontributor(self, name: str, email: str | None = None) -> None:
        """Add contribution comments if necessary."""
        header = self.header()
        if not header:
            return
        prelines = []
        contriblines = []
        postlines = []
        contribexists = False
        incontrib = False
        outcontrib = False
        for line in header.getnotes("translator").split("\n"):
            line = line.strip()
            if line == "FIRST AUTHOR <EMAIL@ADDRESS>, YEAR.":
                incontrib = True
                continue
            if author_re.match(line):
                incontrib = True
                contriblines.append(line)
                continue
            if not line and incontrib:
                incontrib = False
                outcontrib = True
            if incontrib:
                contriblines.append(line)
            elif not outcontrib:
                prelines.append(line)
            else:
                postlines.append(line)

        year = time.strftime("%Y")
        contribexists = False
        for i, line in enumerate(contriblines):
            if name in line and (email is None or email in line):
                contribexists = True
                if year in line:
                    break
                # The contributor is there, but not for this year
                if line[-1] == ".":
                    line = line[:-1]
                contriblines[i] = f"{line}, {year}."

        if not contribexists:
            # Add a new contributor
            if email:
                contriblines.append(f"{name} <{email}>, {year}.")
            else:
                contriblines.append(f"{name}, {year}.")

        header.removenotes()
        header.addnote("\n".join(prelines))
        header.addnote("\n".join(contriblines))
        header.addnote("\n".join(postlines))

    def makeheader(self, **kwargs):
        """
        Create a header for the given filename.

        Check .makeheaderdict() for information on parameters.
        """
        headerpo = self.UnitClass("", encoding=self._encoding)  # ty:ignore[unresolved-attribute]
        # This is hack to make sure proper newlines are used
        # in markfuzzy -> settypecomment calls.
        headerpo._store = self
        headerpo.markfuzzy()
        headeritems = self.makeheaderdict(**kwargs)
        headervalue = "".join(
            [
                f"{key}: {value}\n"
                for key, value in headeritems.items()
                if value is not None
            ]
        )
        headerpo.target = headervalue
        return headerpo



#
# Copyright 2002-2011 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Classes for the support of Gettext .po and .pot files.

This implementation assumes that cpo is working. This should not be used
directly, but can be used once cpo has been established to work.
"""

# TODO:
# - handle headerless PO files better
# - previous msgid and msgctxt
# - accept only unicodes everywhere

import copy
import logging
import re

from translate.misc.multistring import multistring
from translate.storage import base, cpo, pocommon

logger = logging.getLogger(__name__)


lsep = " "
"""Separator for #: entries"""

basic_header = r"""msgid ""
msgstr ""
"Content-Type: text/plain; charset=UTF-8\n"
"Content-Transfer-Encoding: 8bit\n"
"""


class pounit(pocommon.pounit):
    # othercomments = []      #   # this is another comment
    # automaticcomments = []  #   #. comment extracted from the source code
    # sourcecomments = []     #   #: sourcefile.xxx:35
    # prev_msgctxt = []       #   #| The previous values that msgctxt and msgid held
    # prev_msgid = []         #
    # prev_msgid_plural = []  #
    # typecomments = []       #   #, fuzzy
    # msgidcomment = ""      #   _: within msgid
    # msgctxt
    # msgid = []
    # msgstr = []

    # Our homegrown way to indicate what must be copied in a shallow
    # fashion
    __shallow__ = ["_store"]

    def __init__(self, source=None, **kwargs) -> None:
        super().__init__(source)
        self._initallcomments(blankall=True)
        self._msgctxt = ""

        self.target = ""

    def _initallcomments(self, blankall=False) -> None:
        """Initialises allcomments."""
        if blankall:
            self.othercomments = []
            self.automaticcomments = []
            self.sourcecomments = []
            self.typecomments = []
            self.msgidcomment = ""

    @property
    def source(self):
        return self._source

    @source.setter
    def source(self, source) -> None:
        self._rich_source = None
        source = source or ""
        if isinstance(source, (multistring, str)):
            self._source = source
        else:  # If it is unicode, list or dict.
            self._source = multistring(source)

    @property
    def target(self):
        """Returns the unescaped msgstr."""
        return self._target

    @target.setter
    def target(self, target) -> None:
        """Sets the msgstr to the given (unescaped) value."""
        self._rich_target = None
        if self.hasplural():
            if isinstance(target, multistring):
                self._target = target
            else:  # If it is unicode, list or dict.
                self._target = multistring(target)
        elif isinstance(target, (dict, list)):
            if len(target) == 1:
                self._target = target[0]
            else:
                raise ValueError(
                    f"po msgid element has no plural but msgstr has {len(target)} elements ({target})"
                )
        else:
            self._target = target

    def getnotes(self, origin=None):
        """Return comments based on origin value (programmer, developer, source code and translator)."""
        if origin is None:
            comments = "\n".join(self.othercomments)
            comments += "\n".join(self.automaticcomments)
        elif origin == "translator":
            comments = "\n".join(self.othercomments)
        elif origin in {"programmer", "developer", "source code"}:
            comments = "\n".join(self.automaticcomments)
        else:
            raise ValueError("Comment type not valid")
        return comments

    def addnote(self, text, origin=None, position="append") -> None:
        """This is modeled on the XLIFF method. See xliff.py::xliffunit.addnote."""
        # ignore empty strings and strings without non-space characters
        if not (text and text.strip()):
            return
        commentlist = self.othercomments
        autocomments = False
        if origin in {"programmer", "developer", "source code"}:
            autocomments = True
            commentlist = self.automaticcomments
        text = text.removesuffix("\n")
        newcomments = text.split("\n")
        if position == "append":
            newcomments = commentlist + newcomments
        elif position == "prepend":
            newcomments += commentlist

        if autocomments:
            self.automaticcomments = newcomments
        else:
            self.othercomments = newcomments

    def removenotes(self, origin=None) -> None:
        """Remove all the translator's notes (other comments)."""
        self.othercomments = []

    def __deepcopy__(self, memo={}):
        # Make an instance to serve as the copy
        new_unit = self.__class__()
        # We'll be testing membership frequently, so make a set from
        # self.__shallow__
        shallow = set(self.__shallow__)
        # Make deep copies of all members which are not in shallow
        for key, value in self.__dict__.items():
            if key not in shallow:
                setattr(new_unit, key, copy.deepcopy(value))
        # Make shallow copies of all members which are in shallow
        for key in set(shallow):
            setattr(new_unit, key, getattr(self, key))
        # Mark memo with ourself, so that we won't get deep copied
        # again
        memo[id(self)] = self
        # Return our copied unit
        return new_unit

    def copy(self):
        return copy.deepcopy(self)

    def _msgidlen(self):
        if self.hasplural():
            return len("".join(string for string in self.source.strings))
        return len(self.source)

    def _msgstrlen(self):
        if self.hasplural():
            return len("".join(string for string in self.target.strings))
        return len(self.target)

    def merge(
        self, otherpo, overwrite=False, comments=True, authoritative=False
    ) -> None:  # ty:ignore[invalid-method-override]
        """
        Merges the otherpo (with the same msgid) into this one.

        Overwrite non-blank self.msgstr only if overwrite is True
        merge comments only if comments is True
        """

        def mergelists(list1, list2, split=False) -> None:
            # Determine the newline style of list2
            lineend = ""
            if list2 and list2[0]:
                for candidate in ["\n", "\r", "\n\r"]:
                    if list2[0].endswith(candidate):
                        lineend = candidate
                if not lineend:
                    lineend = ""

            # Split if directed to do so:
            if split:
                splitlist1 = []
                splitlist2 = []
                for item in list1:
                    splitlist1.extend(item.split())
                for item in list2:
                    splitlist2.extend(item.split())
                list1.extend([item for item in splitlist2 if item not in splitlist1])
            elif list1 != list2:
                # Normal merge, but conform to list1 newline style
                for item in list2:
                    item = item.rstrip(lineend)
                    # avoid duplicate comment lines (this might cause some problems)
                    if item not in list1 or len(item) < 5:
                        list1.append(item)

        if not isinstance(otherpo, pounit):
            super().merge(otherpo, overwrite, comments)
            return
        if comments:
            mergelists(self.othercomments, otherpo.othercomments)
            mergelists(self.typecomments, otherpo.typecomments)
            if not authoritative:
                # We don't bring across otherpo.automaticcomments as we consider ourself
                # to be the the authority.  Same applies to otherpo.msgidcomments
                mergelists(self.automaticcomments, otherpo.automaticcomments)
                #                mergelists(self.msgidcomments, otherpo.msgidcomments) #XXX?
                mergelists(self.sourcecomments, otherpo.sourcecomments, split=True)
        if not self.istranslated() or overwrite:
            # Remove kde-style comments from the translation (if any). XXX - remove
            if pocommon.extract_msgid_comment(otherpo.target):
                otherpo.target = otherpo.target.replace(
                    f"_: {otherpo._extract_msgidcomments()}\n", ""
                )
            self.target = otherpo.target
            if (
                self.source != otherpo.source
                or self.getcontext() != otherpo.getcontext()
            ):
                self.markfuzzy()
            else:
                self.markfuzzy(otherpo.isfuzzy())
        elif not otherpo.istranslated():
            if self.source != otherpo.source:
                self.markfuzzy()
        elif self.target != otherpo.target:
            self.markfuzzy()

    def isheader(self):
        # TODO: fix up nicely
        return not self.getid() and len(self.target) > 0

    def isblank(self):
        return (
            not self.isheader()
            and not self.msgidcomment
            and (self._msgidlen() == 0)
            and (self._msgstrlen() == 0)
            and len(self._msgctxt) == 0
        )

    def hastypecomment(self, typecomment):
        """Check whether the given type comment is present."""
        # check for word boundaries properly by using a regular expression...
        return (
            sum(
                len(re.findall(f"\\b{typecomment}\\b", tcline))
                for tcline in self.typecomments
            )
            != 0
        )

    def hasmarkedcomment(self, commentmarker):
        """Check whether the given comment marker is present as # (commentmarker) ..."""
        commentmarker = f"({commentmarker})"
        return any(comment.startswith(commentmarker) for comment in self.othercomments)

    def settypecomment(self, typecomment, present=True) -> None:
        """Alters whether a given typecomment is present."""
        if self.hastypecomment(typecomment) != present:
            if present:
                self.typecomments.append(f"#, {typecomment}\n")  # ty:ignore[unresolved-attribute]
            else:
                # this should handle word boundaries properly ...
                typecomments = [
                    re.sub(f"\\b{typecomment}\\b[ \t,]*", "", tcline)
                    for tcline in self.typecomments
                ]
                self.typecomments = filter(
                    lambda tcline: tcline.strip() != "#,", typecomments
                )

    def istranslated(self):
        return super().istranslated() and not self.isobsolete()

    def istranslatable(self) -> bool:
        return not (self.isheader() or self.isblank() or self.isobsolete())

    def isfuzzy(self):
        return self.hastypecomment("fuzzy")

    def _domarkfuzzy(self, present=True) -> None:
        self.settypecomment("fuzzy", present)

    def makeobsolete(self) -> None:
        """Makes this unit obsolete."""
        self.sourcecomments = []
        self.automaticcomments = []
        super().makeobsolete()

    def hasplural(self):
        """Returns whether this pounit contains plural strings..."""
        source = self.source
        return isinstance(source, multistring) and len(source.strings) > 1

    def __str__(self) -> str:
        """Convert to a string. double check that unicode is handled somehow here."""
        cpo_unit = cpo.pounit.buildfromunit(self)
        return str(cpo_unit)

    def getlocations(self):
        """
        Get a list of locations from sourcecomments in the PO unit.

        rtype: List
        return: A list of the locations with '#: ' stripped

        """
        # TODO: rename to .locations
        return self.sourcecomments

    def addlocation(self, location: str) -> None:
        """
        Add a location to sourcecomments in the PO unit.

        :param location: Text location e.g. 'file.c:23' does not include #:
        """
        self.sourcecomments.append(location)

    def _extract_msgidcomments(self, text: str | None = None) -> str:
        """
        Extract KDE style msgid comments from the unit.

        :return: Returns the extracted msgidcomments found in this unit's msgid.
        """
        if text:
            return pocommon.extract_msgid_comment(text)
        return self.msgidcomment

    def getcontext(self):
        """Get the message context."""
        return self._msgctxt + self.msgidcomment

    def setcontext(self, context) -> None:
        self._msgctxt = context or ""

    def getid(self):
        """Returns a unique identifier for this unit."""
        context = self.getcontext()
        # Gettext does not consider the plural to determine duplicates, only
        # the msgid. For generation of .mo files, we might want to use this
        # code to generate the entry for the hash table, but for now, it is
        # commented out for conformance to gettext.
        #        id = '\0'.join(self.source.strings)
        id = self.source
        if self.msgidcomment:
            id = f"_: {context}\n{id}"
        elif context:
            id = f"{context}\04{id}"
        return id

    @classmethod
    def buildfromunit(cls, unit):
        """
        Build a native unit from a foreign unit, preserving as much
        information as possible.
        """
        if type(unit) is cls and hasattr(unit, "copy") and callable(unit.copy):
            return unit.copy()
        if isinstance(unit, pocommon.pounit):
            newunit = cls(unit.source)
            newunit.target = unit.target
            # context
            newunit.msgidcomment = unit._extract_msgidcomments()
            if not newunit.msgidcomment:
                newunit.setcontext(unit.getcontext())

            locations = unit.getlocations()
            if locations:
                newunit.addlocations(locations)
            notes = unit.getnotes("developer")
            if notes:
                newunit.addnote(notes, "developer")
            notes = unit.getnotes("translator")
            if notes:
                newunit.addnote(notes, "translator")
            newunit.markfuzzy(unit.isfuzzy())
            if unit.isobsolete():
                newunit.makeobsolete()
            for tc in ["python-format", "c-format", "php-format"]:
                if unit.hastypecomment(tc):
                    newunit.settypecomment(tc)
                    break
            return newunit
        return base.TranslationUnit.buildfromunit(unit)


class pofile(pocommon.pofile):
    """A .po file containing various units."""

    UnitClass = pounit

    def _build_self_from_cpo(self) -> None:
        """
        Builds up this store from the internal cpo store.

        A user must ensure that self._cpo_store already exists, and that it is
        deleted afterwards.
        """
        for unit in self._cpo_store.units:
            self.addunit(self.UnitClass.buildfromunit(unit))
        self.encoding = self._cpo_store.encoding

    def _build_cpo_from_self(self) -> None:
        """
        Builds the internal cpo store from the data in self.

        A user must ensure that self._cpo_store does not exist, and should
        delete it after using it.
        """
        self._cpo_store = cpo.pofile(noheader=True)
        for unit in self.units:
            if not unit.isblank():
                self._cpo_store.addunit(
                    cpo.pofile.UnitClass.buildfromunit(unit, self.encoding)
                )
        if not self._cpo_store.header():
            # only add a temporary header
            self._cpo_store.makeheader(charset=self.encoding, encoding="8bit")

    def parse(self, input) -> None:  # ty:ignore[invalid-method-override]
        """Parses the given file or file source string."""
        try:
            if hasattr(input, "name"):
                self.filename = input.name
            elif not getattr(self, "filename", ""):
                self.filename = ""
            self.units = []
            self._cpo_store = cpo.pofile(input, noheader=True)
            self._build_self_from_cpo()
            del self._cpo_store
        except Exception as e:
            raise base.ParseError(e) from e

    def removeduplicates(self, duplicatestyle="merge") -> None:
        """Make sure each msgid is unique ; merge comments etc from duplicates into original."""
        # TODO: can we handle consecutive calls to removeduplicates()? What
        # about files already containing msgctxt? - test
        id_dict = {}
        uniqueunits = []
        # TODO: this is using a list as the pos aren't hashable, but this is slow.
        # probably not used frequently enough to worry about it, though.
        markedpos = []

        def addcomment(thepo) -> None:
            thepo.msgidcomment = " ".join(thepo.getlocations())
            markedpos.append(thepo)

        for thepo in self.units:
            id = thepo.getid()
            if thepo.isheader() and not thepo.getlocations():
                # header msgids shouldn't be merged...
                uniqueunits.append(thepo)
            elif id in id_dict:
                if duplicatestyle == "merge":
                    if id:
                        id_dict[id].merge(thepo)
                    else:
                        addcomment(thepo)
                        uniqueunits.append(thepo)
                elif duplicatestyle == "msgctxt":
                    origpo = id_dict[id]
                    if origpo not in markedpos and id:
                        # if it doesn't have an id, we already added msgctxt
                        origpo._msgctxt += " ".join(origpo.getlocations())
                        markedpos.append(thepo)
                    thepo._msgctxt += " ".join(thepo.getlocations())
                    if thepo._msgctxt != id_dict[id]._msgctxt:
                        uniqueunits.append(thepo)
                    else:
                        logger.warning(
                            "Duplicate unit found with msgctx of '%s' and source '%s'",
                            thepo._msgctxt,
                            thepo.source,
                        )
            else:
                if not id:
                    if duplicatestyle == "merge":
                        addcomment(thepo)
                    else:
                        thepo._msgctxt += " ".join(thepo.getlocations())
                id_dict[id] = thepo
                uniqueunits.append(thepo)
        self.units = uniqueunits

    def serialize(self, out) -> None:
        """Write content to file."""
        self._cpo_store = cpo.pofile(encoding=self.encoding, noheader=True)
        try:
            self._build_cpo_from_self()
        except UnicodeEncodeError:
            self.encoding = "utf-8"
            self.updateheader(add=True, Content_Type="text/plain; charset=UTF-8")
            self._build_cpo_from_self()
        self._cpo_store.serialize(out)
        del self._cpo_store



#
# Copyright 2007, 2010 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert JSON files to Gettext PO localization files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/json2po.html
for examples and usage instructions.
"""

import logging

from translate.convert import convert
from translate.storage import jsonl10n, po

logger = logging.getLogger(__name__)


class json2po:
    """Convert a JSON file to a PO file."""

    def convert_store(self, input_store, duplicatestyle="msgctxt"):
        """Converts a JSON file to a PO file."""
        output_store = po.pofile()
        output_header = output_store.header()
        output_header.addnote(f"extracted from {input_store.filename}", "developer")
        for input_unit in input_store.units:
            output_unit = self.convert_unit(input_unit, "developer")
            if output_unit is not None:
                output_store.addunit(output_unit)
        output_store.removeduplicates(duplicatestyle)
        return output_store

    def merge_store(
        self, template_store, input_store, blankmsgstr=False, duplicatestyle="msgctxt"
    ):
        """Converts two JSON files to a PO file."""
        output_store = po.pofile()
        output_header = output_store.header()
        output_header.addnote(
            f"extracted from {template_store.filename}, {input_store.filename}",
            "developer",
        )

        input_store.makeindex()
        for template_unit in template_store.units:
            origpo = self.convert_unit(template_unit, "developer")
            # try and find a translation of the same name...
            template_unit_name = "".join(template_unit.getlocations())
            if template_unit_name in input_store.locationindex:
                translatedjson = input_store.locationindex[template_unit_name]
                translatedpo = self.convert_unit(translatedjson, "translator")
            else:
                translatedpo = None
            # if we have a valid po unit, get the translation and add it...
            if origpo is not None:
                if translatedpo is not None and not blankmsgstr:
                    origpo.target = translatedpo.source
                output_store.addunit(origpo)
            elif translatedpo is not None:
                logger.error(
                    "error converting original JSON definition %s", origpo.name
                )
        output_store.removeduplicates(duplicatestyle)
        return output_store

    @staticmethod
    def convert_unit(input_unit, commenttype):
        """
        Converts a JSON unit to a PO unit.

        :return: None if empty or not for translation
        """
        if input_unit is None:
            return None
        # escape unicode
        output_unit = po.pounit(encoding="UTF-8")
        output_unit.addlocation(input_unit.getid())
        output_unit.source = input_unit.source
        output_unit.target = ""
        return output_unit


def convertjson(
    input_file,
    output_file,
    template_file,
    pot=False,
    duplicatestyle="msgctxt",
    dialect="default",
    filter=None,
) -> int:
    """
    Reads in *input_file* using jsonl10n, converts using :class:`json2po`,
    writes to *output_file*.
    """
    if filter is not None:
        filter = filter.split(",")
    input_store = jsonl10n.JsonFile(input_file, filter=filter)
    convertor = json2po()
    if template_file is None:
        output_store = convertor.convert_store(
            input_store, duplicatestyle=duplicatestyle
        )
    else:
        template_store = jsonl10n.JsonFile(template_file)
        output_store = convertor.merge_store(
            template_store, input_store, blankmsgstr=pot, duplicatestyle=duplicatestyle
        )
    if output_store.isempty():
        return 0
    output_store.serialize(output_file)
    return 1


def main(argv=None) -> None:
    formats = {
        "json": ("po", convertjson),
        ("json", "json"): ("po", convertjson),
    }
    parser = convert.ConvertOptionParser(
        formats, usetemplates=True, usepots=True, description=__doc__
    )
    parser.add_option(
        "",
        "--filter",
        dest="filter",
        default=None,
        help="leaves to extract e.g. 'name,desc': (default: extract everything)",
        metavar="FILTER",
    )
    parser.add_duplicates_option()
    parser.passthrough.append("pot")
    parser.passthrough.append("filter")
    parser.run(argv)


if __name__ == "__main__":
    main()



"""
This module represents the Toki Pona language.

.. seealso:: :wp:`Toki_Pona`
"""

from translate.lang import common


class tok(common.Common):
    """This class represents Toki Pona."""

    ignoretests = {"all": ["simplecaps", "startcaps"]}



#
# Copyright 2009 Zuza Software Foundation
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
This file contains abstract (semantic) interfaces for placeable
 implementations.
"""

from __future__ import annotations

from translate.storage.placeables.strelem import StringElem


class BasePlaceable(StringElem):
    """Base class for all placeables."""


class InvisiblePlaceable(BasePlaceable):
    pass


class MaskingPlaceable(BasePlaceable):
    pass


class ReplacementPlaceable(BasePlaceable):
    pass


class SubflowPlaceable(BasePlaceable):
    pass


class Delimiter:
    pass


class PairedDelimiter:
    pass



#
# Copyright 2008 Zuza Software Foundation
#
# This file is part of translate.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Fill localization files with suggested translations based on
translation memory and existing translations.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/pretranslate.html
for examples and usage instructions.
"""

from translate.convert import convert
from translate.search import match
from translate.storage import factory

# We don't want to reinitialise the TM each time, so let's store it here.
tmmatcher = None


def memory(tmfiles, max_candidates=1, min_similarity=75, max_length=1000):
    """Returns the TM store to use. Only initialises on first call."""
    global tmmatcher  # noqa: PLW0603
    # Only initialise first time
    if tmmatcher is None:
        if isinstance(tmfiles, list):
            tmstore = [factory.getobject(tmfile) for tmfile in tmfiles]
        else:
            tmstore = factory.getobject(tmfiles)
        tmmatcher = match.matcher(
            tmstore,
            max_candidates=max_candidates,
            min_similarity=min_similarity,
            max_length=max_length,
        )
    return tmmatcher


def pretranslate_file(
    input_file,
    output_file,
    template_file,
    tm=None,
    min_similarity=75,
    fuzzymatching=True,
) -> int:
    """
    Pretranslate any factory supported file with old translations and
    translation memory.
    """
    input_store = factory.getobject(input_file)
    template_store = None
    if template_file is not None:
        template_store = factory.getobject(template_file)

    output = pretranslate_store(
        input_store, template_store, tm, min_similarity, fuzzymatching
    )
    output.serialize(output_file)
    return 1


def match_template_location(input_unit, template_store):
    """Returns a matching unit from a template. matching based on locations."""
    # we want to use slightly different matching strategies for PO files
    # generated by our own moz2po and oo2po. Let's take a cheap shot at
    # detecting them from the presence of a ':' in the first location.
    locations = input_unit.getlocations()
    if not locations or ":" in locations[0]:
        return match_template_id(input_unit, template_store)

    # since oo2po and moz2po use location as unique identifiers for strings
    # we match against location first, then check for matching source strings
    # this makes no sense for normal gettext files
    for location in locations:
        matching_unit = template_store.locationindex.get(location, None)
        if (
            matching_unit is not None
            and matching_unit.source == input_unit.source
            and matching_unit.gettargetlen() > 0
        ):
            return matching_unit
    return None


def match_template_id(input_unit, template_store):
    """Returns a matching unit from a template. matching based on unit id."""
    return template_store.findid(input_unit.getid())


def match_source(input_unit, template_store):
    """Returns a matching unit from a template. matching based on unit id."""
    # hack for weird mozilla single letter strings, we don't want to
    # match them by anything but locations
    if len(input_unit.source) > 1:
        return template_store.findunit(input_unit.source)
    return None


def match_fuzzy(input_unit, matchers):
    """Return a fuzzy match from a queue of matchers."""
    for matcher in matchers:
        fuzzycandidates = matcher.matches(input_unit.source)
        if fuzzycandidates:
            return fuzzycandidates[0]
    return None


def pretranslate_unit(
    input_unit, template_store, matchers=None, mark_reused=False, merge_on="id"
):
    """
    Pretranslate a unit or return unchanged if no translation was found.

    :param input_unit: Unit that will be pretranslated.
    :param template_store: Fill input unit with units matching in this store.
    :param matchers: List of fuzzy :class:`~translate.search.match.matcher`
        objects.
    :param mark_reused: Whether to mark old translations as reused or not.
    :param merge_on: Where will the merge matching happen on.
    """
    matching_unit = None

    # Do template matching
    if template_store:
        # :param:`merge_on` supports `location` and `id` for now
        if merge_on == "location":
            matching_unit = match_template_location(input_unit, template_store)
        else:
            matching_unit = match_template_id(input_unit, template_store)

    if matching_unit and matching_unit.gettargetlen() > 0:
        input_unit.merge(matching_unit, authoritative=True)
    elif matchers:
        # quickly try exact match by source
        matching_unit = match_source(input_unit, template_store)

        if not matching_unit or not matching_unit.gettargetlen():
            # do fuzzy matching
            matching_unit = match_fuzzy(input_unit, matchers)

        if matching_unit and matching_unit.gettargetlen() > 0:
            # FIXME: should we dispatch here instead of this crude attr check
            if hasattr(input_unit, "addalttrans"):
                # FIXME: what about origin, lang and matchquality
                input_unit.addalttrans(
                    matching_unit.target, origin="fish", sourcetxt=matching_unit.source
                )
            else:
                input_unit.merge(matching_unit, authoritative=True)

    # FIXME: ugly hack required by pot2po to mark old
    # translations reused for new file. loops over
    if mark_reused and matching_unit and template_store:
        original_unit = template_store.findunit(matching_unit.source)
        if original_unit is not None:
            original_unit.reused = True

    return input_unit


def pretranslate_store(
    input_store, template_store, tm=None, min_similarity=75, fuzzymatching=True
):
    """Do the actual pretranslation of a whole store."""
    # preparation
    matchers = []
    # prepare template
    if template_store is not None:
        template_store.makeindex()
        # template preparation based on type
        prepare_template = f"prepare_template_{template_store.__class__.__name__}"
        if prepare_template in globals():
            globals()[prepare_template](template_store)

        if fuzzymatching:
            # create template matcher
            # FIXME: max_length hardcoded
            matcher = match.matcher(
                template_store,
                max_candidates=1,
                min_similarity=min_similarity,
                max_length=3000,
                usefuzzy=True,
            )
            matcher.addpercentage = False
            matchers.append(matcher)

    # prepare tm
    # create tm matcher
    if tm and fuzzymatching:
        # FIXME: max_length hardcoded
        matcher = memory(
            tm, max_candidates=1, min_similarity=min_similarity, max_length=1000
        )
        matcher.addpercentage = False
        matchers.append(matcher)

    # Main loop
    for input_unit in input_store.units:
        if input_unit.istranslatable():
            pretranslate_unit(
                input_unit, template_store, matchers, merge_on=input_store.merge_on
            )

    return input_store


def main(argv=None) -> None:
    formats = {
        "pot": ("po", pretranslate_file),
        ("pot", "po"): ("po", pretranslate_file),
        "po": ("po", pretranslate_file),
        ("po", "po"): ("po", pretranslate_file),
        "xlf": ("xlf", pretranslate_file),
        ("xlf", "xlf"): ("xlf", pretranslate_file),
        "xliff": ("xliff", pretranslate_file),
        ("xliff", "xliff"): ("xliff", pretranslate_file),
    }
    parser = convert.ConvertOptionParser(
        formats, usetemplates=True, allowmissingtemplate=True, description=__doc__
    )
    parser.add_option(
        "",
        "--tm",
        dest="tm",
        default=None,
        help="The file to use as translation memory when fuzzy matching",
    )
    parser.passthrough.append("tm")
    defaultsimilarity = 75
    parser.add_option(
        "-s",
        "--similarity",
        dest="min_similarity",
        default=defaultsimilarity,
        type="float",
        help=f"The minimum similarity for inclusion (default: {defaultsimilarity}%)",
    )
    parser.passthrough.append("min_similarity")
    parser.add_option(
        "--nofuzzymatching",
        dest="fuzzymatching",
        action="store_false",
        default=True,
        help="Disable fuzzy matching",
    )
    parser.passthrough.append("fuzzymatching")
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2025 Translate toolkit contributors
#
# This file is part of the Translate Toolkit.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Common functionality shared between XLIFF 1.x and XLIFF 2.0 implementations."""

from typing import Any, TypeVar

from lxml import etree

from translate.misc.multistring import multistring
from translate.misc.xml_helpers import (
    clear_content,
    getXMLspace,
    safely_set_text,
    setXMLspace,
)
from translate.storage import lisa
from translate.storage.placeables.lisa import strelem_to_xml, xml_to_strelem


class XliffUnit(lisa.LISAunit):
    """Base class providing common functionality for XLIFF units."""

    def _ensure_xml_space_preserve(self) -> None:
        """Ensure xml:space='preserve' is set on the unit."""
        if getXMLspace(self.xmlelement) != "preserve":
            setXMLspace(self.xmlelement, "preserve")

    def createlanguageNode(self, lang: str, text: str, purpose: str) -> etree._Element:  # ty:ignore[invalid-method-override]
        """
        Create and return an xml Element setup with given parameters.

        This is a common implementation that works for both XLIFF 1.x and 2.0.
        """
        assert purpose
        langset = etree.Element(self.namespaced(purpose))
        safely_set_text(langset, text)
        return langset

    @classmethod
    def multistring_to_rich(cls, mstr: Any) -> list[Any]:  # ty:ignore[invalid-method-override]
        """
        Convert a multistring to rich format.

        Override :meth:`TranslationUnit.multistring_to_rich` which is used
        by the ``rich_source`` and ``rich_target`` properties.
        """
        strings = mstr
        if isinstance(mstr, multistring):
            strings = mstr.strings
        elif isinstance(mstr, str):
            strings = [mstr]
        return [xml_to_strelem(s) for s in strings]

    @classmethod
    def rich_to_multistring(cls, elem_list: list[Any]) -> multistring:
        """
        Convert rich format to multistring.

        Override :meth:`TranslationUnit.rich_to_multistring` which is used
        by the ``rich_source`` and ``rich_target`` properties.
        """
        return multistring([str(elem) for elem in elem_list])

    def getlanguageNodes(self):
        """We override this to get source and target nodes."""
        source = None
        target = None
        nodes = []
        try:
            source = next(
                self.xmlelement.iterchildren(self.namespaced(self.languageNode))
            )
            target = next(self.xmlelement.iterchildren(self.namespaced("target")))
            nodes = [source, target]
        except StopIteration:
            if source is not None:
                nodes.append(source)
            if target is not None:
                nodes.append(target)
        return nodes

    def set_rich_source(self, value, sourcelang="en") -> None:
        sourcelanguageNode = self.get_source_dom()
        if sourcelanguageNode is None:
            sourcelanguageNode = self.createlanguageNode(sourcelang, "", "source")
            self.set_source_dom(sourcelanguageNode)

        # Clear sourcelanguageNode first
        clear_content(sourcelanguageNode)

        strelem_to_xml(sourcelanguageNode, value[0])

    @property
    def rich_source(self):
        # rsrc = xml_to_strelem(self.source_dom)
        # logger.debug('rich source: %s' % (repr(rsrc)))
        # from dubulib.debug.misc import print_stack_funcs
        # print_stack_funcs()
        return [
            xml_to_strelem(
                self.source_dom, getXMLspace(self.xmlelement, self._default_xml_space)
            )
        ]

    @rich_source.setter
    def rich_source(self, value) -> None:
        self.set_rich_source(value)

    def set_rich_target(self, value, lang="xx", append=False) -> None:
        self._rich_target = None
        if value is None:
            self.set_target_dom(self.createlanguageNode(lang, "", "target"))
            return

        self._ensure_xml_space_preserve()
        languageNode = self.get_target_dom()
        if languageNode is None:
            languageNode = self.createlanguageNode(lang, "", "target")
            self.set_target_dom(languageNode, append)

        # Clear languageNode first
        clear_content(languageNode)

        strelem_to_xml(languageNode, value[0])
        ### currently giving some issues in Virtaal: self._rich_target = value

    def get_rich_target(self, lang=None):
        """
        Retrieves the "target" text (second entry), or the entry in the
        specified language, if it exists.
        """
        if self._rich_target is None:
            self._rich_target = [
                xml_to_strelem(
                    self.get_target_dom(lang),
                    getXMLspace(self.xmlelement, self._default_xml_space),
                )
            ]
        return self._rich_target

    @property
    def rich_target(self):
        return self.get_rich_target()

    @rich_target.setter
    def rich_target(self, value) -> None:
        self.set_rich_target(value)

    def merge(
        self, otherunit, overwrite=False, comments=True, authoritative=False
    ) -> None:
        # TODO: consider other attributes like "approved"
        super().merge(otherunit, overwrite, comments)
        if self.target:
            self.marktranslated()  # ty:ignore[unresolved-attribute]
            if otherunit.isfuzzy():
                self.markfuzzy()
            elif otherunit.source == self.source:
                self.markfuzzy(False)
            elif otherunit.source != self.source:
                self.markfuzzy(True)
        if comments:
            self.addnote(otherunit.getnotes())


U = TypeVar("U", bound=XliffUnit)


class XliffFile(lisa.LISAfile[U]):
    """Base class providing common functionality for XLIFF file stores."""

    def __init__(self, *args, **kwargs) -> None:
        """Initialize XLIFF file with filename tracking."""
        self._filename = None
        super().__init__(*args, **kwargs)

    @staticmethod
    def getfilename(filenode) -> str | None:
        """
        Returns the identifier of the given file node.

        Must be overridden by subclasses to specify which attribute to use.
        """
        raise NotImplementedError("Subclasses must implement getfilename()")

    @staticmethod
    def setfilename(filenode, filename: str) -> None:
        """
        Set the identifier of the given file node.

        Must be overridden by subclasses to specify which attribute to use.
        """
        raise NotImplementedError("Subclasses must implement setfilename()")

    def getfilenames(self):
        """Returns all file identifiers in this XLIFF file."""
        filenodes = self.document.getroot().iterchildren(self.namespaced("file"))
        filenames = [self.getfilename(filenode) for filenode in filenodes]
        return list(filter(None, filenames))

    def getfilenode(self, filename, createifmissing=False):
        """Finds the file node with the given identifier."""
        filenodes = self.document.getroot().iterchildren(self.namespaced("file"))
        for filenode in filenodes:
            if self.getfilename(filenode) == filename:
                return filenode
        if not createifmissing:
            return None
        return self.createfilenode(filename)  # ty:ignore[unresolved-attribute]



#
# Copyright 2002-2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert Gettext PO localization files to JSON files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/json2po.html
for examples and usage instructions.
"""

from translate.convert import convert
from translate.storage import factory, jsonl10n


class rejson:
    def __init__(self, templatefile, inputstore) -> None:
        self.templatefile = templatefile
        self.templatestore = jsonl10n.JsonFile(templatefile)
        self.ouputstore = jsonl10n.JsonFile()
        self.inputstore = inputstore

    def convertstore(self, includefuzzy=False, remove_untranslated=False):
        self.includefuzzy = includefuzzy
        self.remove_untranslated = remove_untranslated
        self.inputstore.makeindex()
        for unit in self.templatestore.units:
            inputunit = self.inputstore.locationindex.get(unit.getid())
            skip_unit = self.remove_untranslated and (
                inputunit is None or inputunit.isfuzzy() or not inputunit.istranslated()
            )
            if skip_unit:
                continue
            if inputunit is not None:
                if inputunit.isfuzzy():
                    if self.includefuzzy:
                        # inputunit.istranslated() is always False now,
                        # because inputunit.isfuzzy() is True.
                        # So we need to check if inputunit.target is truthy.
                        if inputunit.target:
                            unit.target = inputunit.target
                        else:
                            unit.target = unit.source
                    else:
                        unit.target = unit.source
                elif inputunit.istranslated():
                    unit.target = inputunit.target
                else:
                    unit.target = unit.source
            else:
                unit.target = unit.source
            self.ouputstore.addunit(unit)
        return bytes(self.ouputstore)


def convertjson(
    inputfile,
    outputfile,
    templatefile,
    includefuzzy=False,
    outputthreshold=None,
    remove_untranslated=False,
) -> bool:
    inputstore = factory.getobject(inputfile)

    if not convert.should_output_store(inputstore, outputthreshold):
        return False

    if templatefile is None:
        raise ValueError("Must have template file for JSON files")

    convertor = rejson(templatefile, inputstore)
    outputstring = convertor.convertstore(includefuzzy, remove_untranslated)
    outputfile.write(outputstring)
    return True


def main(argv=None) -> None:
    # handle command line options
    formats = {
        ("po", "json"): ("json", convertjson),
    }
    parser = convert.ConvertOptionParser(
        formats, usetemplates=True, description=__doc__
    )
    parser.add_threshold_option()
    parser.add_fuzzy_option()
    parser.add_remove_untranslated_option()
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2018 BhaaL
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""Module for handling flat XML files."""

from lxml import etree

from translate.misc.xml_helpers import getText, namespaced, reindent, safely_set_text
from translate.storage import base


class FlatXMLUnit(base.TranslationUnit):
    """A single term in the XML file."""

    DEFAULT_NAMESPACE = None
    DEFAULT_ELEMENT_NAME = "str"
    DEFAULT_ATTRIBUTE_NAME = "key"

    def __init__(
        self,
        source=None,
        namespace=None,
        element_name=None,
        attribute_name=None,
        **kwargs,
    ) -> None:
        self.namespace = namespace or self.DEFAULT_NAMESPACE
        self.element_name = element_name or self.DEFAULT_ELEMENT_NAME
        self.attribute_name = attribute_name or self.DEFAULT_ATTRIBUTE_NAME
        self.xmlelement = etree.Element(self.namespaced(self.element_name))
        super().__init__(source, **kwargs)

    def __str__(self) -> str:
        # "unicode" encoding keeps the unicode status of the output
        return etree.tostring(self.xmlelement, encoding="unicode")

    @property
    def source(self):
        """Returns the unique identifier of this unit."""
        return self.xmlelement.get(self.attribute_name)

    @source.setter
    def source(self, source) -> None:
        """Updates the unique identifier of this unit."""
        self.xmlelement.set(self.attribute_name, source)

    @property
    def target(self):
        """Returns the translated string of this unit."""
        return self.node_text

    @target.setter
    def target(self, target) -> None:
        """Updates the translated string of this unit."""
        if self.target == target:
            return
        safely_set_text(self.xmlelement, target)

    def namespaced(self, name):
        """Returns name in Clark notation."""
        return namespaced(self.namespace, name)

    @property
    def node_text(self):
        """Returns the text content of the XML element."""
        if self.xmlelement is None:
            return None

        return getText(self.xmlelement)

    @classmethod
    def createfromxmlElement(
        cls, element, namespace=None, element_name="str", attribute_name="key"
    ):
        """
        Attempts to create a unit from the passed element.

        element must not be None and must match the given element name
        (including namespace); otherwise None will be returned.
        """
        if element is None:
            return None
        if element.tag != namespaced(namespace, element_name):
            return None
        unit = cls(
            source=None,
            namespace=namespace,
            element_name=element_name,
            attribute_name=attribute_name,
        )
        unit.xmlelement = element
        return unit


class NOTPROVIDED:
    pass


class FlatXMLFile(base.TranslationStore):
    """Class representing a flat XML file store."""

    UnitClass = FlatXMLUnit
    _name = "Flat XML File"
    Mimetypes = ["text/xml"]
    Extensions = ["xml"]

    DEFAULT_ROOT_NAME = "root"
    DEFAULT_VALUE_NAME = "str"
    DEFAULT_KEY_NAME = "key"
    DEFAULT_NAMESPACE = None
    DEFAULT_INDENT_CHARS = "  "
    DEFAULT_TRAILING_EOL = True
    XML_DECLARATION = True

    def __init__(
        self,
        inputfile=None,
        sourcelanguage="en",
        targetlanguage=None,
        root_name=None,
        value_name=None,
        key_name=None,
        namespace=None,
        indent_chars=NOTPROVIDED,
        trailing_eol=None,
        **kwargs,
    ) -> None:
        self.root_name = root_name or self.DEFAULT_ROOT_NAME
        self.value_name = value_name or self.DEFAULT_VALUE_NAME
        self.key_name = key_name or self.DEFAULT_KEY_NAME
        self.namespace = namespace or self.DEFAULT_NAMESPACE
        self.indent_chars = (
            self.DEFAULT_INDENT_CHARS if indent_chars is NOTPROVIDED else indent_chars
        )
        self.trailing_eol = (
            self.DEFAULT_TRAILING_EOL if trailing_eol is None else trailing_eol
        )

        super().__init__(**kwargs)
        if inputfile is not None:
            self.parse(inputfile)
        else:
            self.make_empty_file()
            self.setsourcelanguage(sourcelanguage)
            self.settargetlanguage(targetlanguage)

    def addunit(self, unit, new=True) -> None:
        unit.namespace = self.namespace
        super().addunit(unit)
        if new:
            self.root.append(unit.xmlelement)

    def removeunit(self, unit) -> None:
        super().removeunit(unit)
        self.root.remove(unit.xmlelement)

    def reindent(self) -> None:
        """Reindents the backing document to be consistent."""
        # no elements? nothing to do.
        if not len(self.root):
            pass

        if self.indent_chars is None:
            # indent None means: linearize
            self.root.text = None
            for child in self.root:
                child.tail = None
        else:
            reindent(self.root, indent=self.indent_chars)

        if self.trailing_eol:
            # ensure trailing EOL for VCS
            self.root.tail = "\n"

    def serialize(self, out) -> None:
        self.reindent()
        self.document.write(
            out, xml_declaration=self.XML_DECLARATION, encoding=self.encoding
        )

    def make_empty_file(self) -> None:
        """Initializes the backing document to be an empty root element."""
        self.root = etree.Element(self.namespaced(self.root_name))
        self.document = self.root.getroottree()

    def parse(self, xml) -> None:  # ty:ignore[invalid-method-override]
        """Parses the passed xml file into the backing document."""
        if not hasattr(self, "filename"):
            self.filename = getattr(xml, "name", "")
        if hasattr(xml, "read"):
            xml.seek(0)
            posrc = xml.read()
            xml = posrc

        parser = etree.XMLParser(strip_cdata=False, resolve_entities=False)
        self.root = etree.fromstring(xml, parser)
        self.document = self.root.getroottree()
        self.encoding = self.document.docinfo.encoding

        root_name = self.namespaced(self.root_name)
        assert self.root.tag == root_name, (
            f"expected root name to be {root_name} but got {self.root.tag}"
        )
        if len(self.root):
            # we'd expect at least one child element to have the correct
            # name and attributes; otherwise the name parameters might've
            # been wrong/typo'd and need to be addressed in order to avoid
            # coming up empty when the file actually contains entries.
            value_name = self.namespaced(self.value_name)
            matching_nodes = list(self.root.iterchildren(value_name))
            assert matching_nodes, (
                f"expected value name to be {value_name} but first node is {self.root[0].tag}"
            )

            assert matching_nodes[0].get(self.key_name), (
                f"expected key attribute to be {self.key_name}, found attribute(s): {','.join(matching_nodes[0].attrib)}"
            )

        for entry in self.root:
            unit = self.UnitClass.createfromxmlElement(
                entry,
                namespace=self.namespace,
                element_name=self.value_name,
                attribute_name=self.key_name,
            )
            if unit is not None:
                self.addunit(unit, new=False)

    def namespaced(self, name):
        """Returns name in Clark notation."""
        return namespaced(self.namespace, name)



#
# Copyright 2005, 2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert Gettext PO localization files to XLIFF localization files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/xliff2po.html
for examples and usage instructions.
"""

from translate.convert import convert
from translate.storage import po, poxliff


class po2xliff:
    def convertunit(self, outputstore, inputunit, filename):
        """Creates a transunit node."""
        source = inputunit.source
        target = inputunit.target
        if inputunit.isheader():
            unit = outputstore.addheaderunit(target, filename)
        else:
            unit = outputstore.addsourceunit(source, filename, True)
            unit.target = target
            # Explicitly marking the fuzzy state will ensure that normal (translated)
            # units in the PO file end up as approved in the XLIFF file.
            if target:
                unit.markfuzzy(inputunit.isfuzzy())
            else:
                unit.markapproved(False)

            # Handle #: location comments
            for location in inputunit.getlocations():
                unit.createcontextgroup(
                    "po-reference", self.contextlist(location), purpose="location"
                )

            # Handle #. automatic comments
            comment = inputunit.getnotes("developer")
            if comment:
                unit.createcontextgroup(
                    "po-entry", [("x-po-autocomment", comment)], purpose="information"
                )
                unit.addnote(comment, origin="developer")

            # TODO: x-format, etc.

        # Handle # other comments
        comment = inputunit.getnotes("translator")
        if comment:
            unit.createcontextgroup(
                "po-entry", [("x-po-trancomment", comment)], purpose="information"
            )
            unit.addnote(comment, origin="po-translator")

        return unit

    @staticmethod
    def contextlist(location):
        contexts = []
        if ":" in location:
            sourcefile, linenumber = location.split(":", 1)
        else:
            sourcefile, linenumber = location, None
        contexts.append(("sourcefile", sourcefile))
        if linenumber:
            contexts.append(("linenumber", linenumber))
        return contexts

    def convertstore(self, inputstore, templatefile=None, **kwargs):
        """Converts a .po file to .xlf format."""
        if templatefile is None:
            outputstore = poxliff.PoXliffFile(**kwargs)
        else:
            outputstore = poxliff.PoXliffFile(templatefile, **kwargs)
        filename = inputstore.filename
        for inputunit in inputstore.units:
            if inputunit.isblank():
                continue
            self.convertunit(outputstore, inputunit, filename)
        return bytes(outputstore)


def convertpo(inputfile, outputfile, templatefile) -> int:
    """Reads in stdin using fromfileclass, converts using convertorclass, writes to stdout."""
    inputstore = po.pofile(inputfile)
    if inputstore.isempty():
        return 0
    convertor = po2xliff()
    outputstring = convertor.convertstore(inputstore, templatefile)
    outputfile.write(outputstring)
    return 1


def main(argv=None) -> None:
    formats = (
        ("po", ("xlf", convertpo)),
        (("po", "xlf"), ("xlf", convertpo)),
        ("po", ("xliff", convertpo)),
        (("po", "xliff"), ("xliff", convertpo)),
    )
    parser = convert.ConvertOptionParser(
        formats, usetemplates=True, description=__doc__
    )
    parser.run(argv)


if __name__ == "__main__":
    main()



#
# Copyright 2008, 2014 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

import zipfile


def open_odf(filename):
    z = zipfile.ZipFile(filename, "r")
    return {
        "content.xml": z.read("content.xml"),
        "meta.xml": z.read("meta.xml"),
        "styles.xml": z.read("styles.xml"),
    }


def copy_odf(input_zip, output_zip, exclusion_list):
    for name in [name for name in input_zip.namelist() if name not in exclusion_list]:
        output_zip.writestr(name, input_zip.read(name))
    return output_zip



#
# Copyright 2019 Cademia Siciliana <tech@cademiasiciliana.org>
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
This module represents the Sicilian language.

.. seealso:: :wp:`Sicilian_language`
"""

from translate.filters.checks import CheckerConfig, FilterFailure, TranslationChecker
from translate.filters.decorators import critical
from translate.lang import common


def contains_illegal(illegal_substrings, string):
    """
    Check if string contains any of the specified illegal substrings.

    :param illegal_substrings: an iterable of illegal substrings
    :param string: the string to check against occurrences of illegal substrings
    :return: True if string contains any of the illegal substrings
    """
    return any(s in string for s in illegal_substrings)


sicilianconfig = CheckerConfig(
    validchars="àèìòùḍÀÈÌÒÙḌ",
)


class SicilianChecker(TranslationChecker):
    """A Checker class for Sicilian."""

    def __init__(self, **kwargs) -> None:
        checkerconfig = kwargs.get("checkerconfig")

        if checkerconfig is None:
            checkerconfig = CheckerConfig()
            kwargs["checkerconfig"] = checkerconfig

        checkerconfig.update(sicilianconfig)
        super().__init__(**kwargs)

    @critical
    def italianisms(self, str1, str2) -> bool:
        """
        Check if the translation contains common errors done by italophones.

        Mainly inspired by musttranslatewords(), but with a different logic: return True if
        the given word appears in the translation but not in the source (if it's in the source,
        then presumably it's being kept untranslated).

        :param str1: the source string
        :param str2: the target (translated) string
        :return: True if str2 doesn't contain an "italianism"
        """
        str1 = self.removevariables(str1)
        str2 = self.removevariables(str2)

        errors = {
            "io": "ju/jeu/iu/...",
            "tantu": "assai",
            "menu": "cchiù picca",
        }

        # The above is full of strange quotes and things in utf-8 encoding.
        # single apostrophe perhaps problematic in words like "doesn't"
        for separator in self.config.punctuation:
            str1 = str1.replace(separator, " ")
            str2 = str2.replace(separator, " ")

        words1 = self.filteraccelerators(str1).split()
        words2 = self.filteraccelerators(str2).split()
        stopwords = [
            f"{word} ({errors[word]})"
            for word in words2
            if word.lower() in errors and word not in words1
        ]

        if stopwords:
            raise FilterFailure(f"Please translate: {', '.join(stopwords)}")

        return True

    @critical
    def vocalism(self, str1, str2) -> bool:
        """
        Check correct word-endings.

        All words should end with a/i/u, but a handful of exceptions:

          - me, to, so (possessive pronouns)
          - po (verb "putiri")
          - no
          - jo (personal pronoun)
          - se (yes)
          - nne (in)

        :param str1: the source string
        :param str2: the target (translated) string
        :return: True if there are no words with endings not in respect of vocalism (or if they appear in source string as well)
        """
        exceptions = ["me", "to", "so", "po", "no", "jo", "se", "nne"]

        stopwords = []

        for word in self.config.lang.words(str2):
            lower_word = word.lower()
            if (
                word not in str1
                and lower_word not in exceptions
                and lower_word.endswith(("e", "o"))
                and lower_word not in stopwords
            ):
                stopwords.append(lower_word)

        if stopwords:
            raise FilterFailure(f"Please respect vocalism: {', '.join(stopwords)}")
        return True

    @critical
    def suffixes(self, str1, str2) -> bool:
        """
        Check for common word suffixes to be written correctly.

        :param str1: the source string
        :param str2: the target (translated) string
        :return: True if there are no common suffixes wrongly written
        """
        suffixes = {
            "zzioni": "zziuni",
        }

        stopwords = [
            f"{word} (-{value})"
            for word in self.config.lang.words(str2)
            for suffix, value in suffixes.items()
            if word not in str1 and word.lower().endswith(suffix)
        ]

        if stopwords:
            raise FilterFailure(
                f"Please use the correct word endings: {', '.join(stopwords)}"
            )
        return True


class scn(common.Common):
    """This class represents Sicilian."""

    checker = SicilianChecker()

    ignoretests = {
        "all": ["doublewords"],
    }



#
# Copyright 2004-2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Handles converting of files between formats (used by
:mod:`translate.convert` tools).
"""

import os.path
from io import BytesIO

from translate.misc import optrecurse
from translate.tools import pocount

# Don't import optparse ourselves, get the version from optrecurse.
optparse = optrecurse.optparse


class ConvertOptionParser(optrecurse.RecursiveOptionParser):
    """A specialized Option Parser for convertor tools..."""

    def __init__(
        self,
        formats,
        usetemplates=False,
        usepots=False,
        allowmissingtemplate=False,
        description=None,
    ) -> None:
        """Construct the specialized Option Parser."""
        super().__init__(
            formats,
            usetemplates,
            allowmissingtemplate=allowmissingtemplate,
            description=description,
        )
        self.usepots = usepots
        self.settimestampoption()
        self.setpotoption()
        self.set_usage()

    def add_fuzzy_option(self, default=False) -> None:
        """Adds an option to include / exclude fuzzy translations."""
        fuzzyhelp = "use translations marked fuzzy"
        nofuzzyhelp = "don't use translations marked fuzzy"
        if default:
            fuzzyhelp += " (default)"
        else:
            nofuzzyhelp += " (default)"
        self.add_option(
            "",
            "--fuzzy",
            dest="includefuzzy",
            action="store_true",
            default=default,
            help=fuzzyhelp,
        )
        self.add_option(
            "",
            "--nofuzzy",
            dest="includefuzzy",
            action="store_false",
            default=default,
            help=nofuzzyhelp,
        )
        self.passthrough.append("includefuzzy")

    def add_remove_untranslated_option(self, default=False) -> None:
        """
        Adds an option to remove key value from output if it is
        untranslated.
        """
        self.add_option(
            "",
            "--removeuntranslated",
            dest="remove_untranslated",
            default=False,
            action="store_true",
            help="remove untranslated strings from output",
        )
        self.passthrough.append("remove_untranslated")

    def add_threshold_option(self, default=None) -> None:
        """
        Adds an option to output only stores where translation percentage
        exceeds the threshold.
        """
        self.add_option(
            "",
            "--threshold",
            dest="outputthreshold",
            default=default,
            metavar="PERCENT",
            type="int",
            help="only convert files where the translation completion is above PERCENT",
        )
        self.passthrough.append("outputthreshold")

    def add_duplicates_option(self, default="msgctxt") -> None:
        """Adds an option to say what to do with duplicate strings."""
        self.add_option(
            "",
            "--duplicates",
            dest="duplicatestyle",
            default=default,
            type="choice",
            choices=["msgctxt", "merge"],
            help=f"what to do with duplicate strings (identical source text): merge, msgctxt (default: '{default}')",
            metavar="DUPLICATESTYLE",
        )
        self.passthrough.append("duplicatestyle")

    def add_multifile_option(self, default="single") -> None:
        """Adds an option to say how to split the po/pot files."""
        self.add_option(
            "",
            "--multifile",
            dest="multifilestyle",
            default=default,
            type="choice",
            choices=["single", "toplevel", "onefile"],
            help="how to split po/pot files (single, toplevel or onefile)",
            metavar="MULTIFILESTYLE",
        )
        self.passthrough.append("multifilestyle")

    @staticmethod
    def potifyformat(fileformat):
        """Converts a .po to a .pot where required."""
        if fileformat is None:
            return fileformat
        if fileformat == "po":
            return "pot"
        if fileformat.endswith(f"{os.extsep}po"):
            return f"{fileformat}t"
        return fileformat

    def getformathelp(self, formats):
        """Make a nice help string for describing formats..."""
        # include implicit pot options...
        helpformats = []
        for fileformat in formats:
            helpformats.append(fileformat)
            potformat = self.potifyformat(fileformat)
            if potformat != fileformat:
                helpformats.append(potformat)
        return super().getformathelp(helpformats)

    def filterinputformats(self, options):
        """Filters input formats, processing relevant switches in options."""
        if self.usepots and options.pot:
            return [self.potifyformat(inputformat) for inputformat in self.inputformats]
        return self.inputformats

    def filteroutputoptions(self, options):
        """Filters output options, processing relevant switches in options."""
        if self.usepots and options.pot:
            outputoptions = {}
            for (inputformat, templateformat), (
                outputformat,
                convertor,
            ) in self.outputoptions.items():
                inputformat = self.potifyformat(inputformat)
                templateformat = self.potifyformat(templateformat)
                outputformat = self.potifyformat(outputformat)
                outputoptions[inputformat, templateformat] = (outputformat, convertor)
            return outputoptions
        return self.outputoptions

    def setpotoption(self) -> None:
        """
        Sets the ``-P``/``--pot`` option depending on input/output formats
        etc.
        """
        if self.usepots:
            potoption = optparse.Option(
                "-P",
                "--pot",
                action="store_true",
                dest="pot",
                default=False,
                help="output PO Templates (.pot) rather than PO files (.po)",
            )
            self.define_option(potoption)

    def settimestampoption(self) -> None:
        """Sets ``-S``/``--timestamp`` option."""
        timestampopt = optparse.Option(
            "-S",
            "--timestamp",
            action="store_true",
            dest="timestamp",
            default=False,
            help="skip conversion if the output file has newer timestamp",
        )
        self.define_option(timestampopt)

    def verifyoptions(self, options) -> None:
        """
        Verifies that the options are valid (required options are present,
        etc).
        """

    def run(self, argv=None) -> None:
        """Parses the command line options and runs the conversion."""
        options, _args = self.parse_args(argv)
        self.inputformats = self.filterinputformats(options)
        self.outputoptions = self.filteroutputoptions(options)
        try:
            self.verifyoptions(options)
        except Exception as e:
            self.error(str(e))
        self.recursiveprocess(options)

    def processfile(
        self, fileprocessor, options, fullinputpath, fulloutputpath, fulltemplatepath
    ):
        # The `fulloutputpath`s parsed when `onefile` is enabled don't exist, which causes
        # the `_output_is_newer` check to return False. This means that even if there is
        # no content update, the POT/PO files are regenerated. When `onefile` is enabled,
        # this check sets the output path used by `_output_is_newer` to the output file
        # provided to the command, which will exist. It remains `fulloutputpath` otherwise.
        if getattr(options, "multifilestyle", None) == "onefile":
            timecheck_output_path = options.output
        else:
            timecheck_output_path = fulloutputpath

        if options.timestamp and _output_is_newer(fullinputpath, timecheck_output_path):
            return False

        return super().processfile(
            fileprocessor, options, fullinputpath, fulloutputpath, fulltemplatepath
        )


def copyinput(inputfile, outputfile, templatefile, **kwargs) -> bool:
    """Copies the input file to the output file."""
    outputfile.write(inputfile.read())
    return True


def copytemplate(inputfile, outputfile, templatefile, **kwargs) -> bool:
    """Copies the template file to the output file."""
    outputfile.write(templatefile.read())
    return True


class Replacer:
    """An object that knows how to replace strings in files."""

    def __init__(self, searchstring, replacestring) -> None:
        self.searchstring = searchstring
        self.replacestring = replacestring

    def doreplace(self, text):
        """Actually replace the text."""
        if self.searchstring is not None and self.replacestring is not None:
            return text.replace(self.searchstring, self.replacestring)
        return text

    def searchreplaceinput(self, inputfile, outputfile, templatefile, **kwargs) -> bool:
        """Copies the input file to the output file, searching and replacing."""
        outputfile.write(self.doreplace(inputfile.read()))
        return True

    def searchreplacetemplate(
        self, inputfile, outputfile, templatefile, **kwargs
    ) -> bool:
        """
        Copies the template file to the output file, searching and
        replacing.
        """
        outputfile.write(self.doreplace(templatefile.read()))
        return True


# archive files need to know how to:
# - openarchive: creates an archive object for the archivefilename
#   * requires a constructor that takes the filename
# - iterarchivefile: iterate through the names in the archivefile
#   * requires the default iterator to do this
# - archivefileexists: check if a given pathname exists inside the archivefile
#   * uses the in operator - requires __contains__ (or will use __iter__
#     by default)
# - openarchiveinputfile: returns an open input file from the archive, given
#   the path
#   * requires an archivefile.openinputfile method that takes the pathname
# - openarchiveoutputfile: returns an open output file from the archive, given
#   the path
#   * requires an archivefile.openoutputfile method that takes the pathname


class ArchiveConvertOptionParser(ConvertOptionParser):
    """
    ConvertOptionParser that can handle recursing into single archive files.

    ``archiveformats`` maps extension to class. If the extension doesn't
    matter, it can be None.

    If the extension is only valid for input/output/template, it can be given
    as ``(extension, filepurpose)``.
    """

    def __init__(
        self,
        formats,
        usetemplates=False,
        usepots=False,
        description=None,
        archiveformats=None,
    ) -> None:
        if archiveformats is None:
            self.archiveformats = {}
        else:
            self.archiveformats = archiveformats
        self.archiveoptions = {}
        super().__init__(formats, usetemplates, usepots, description=description)

    def isrecursive(self, fileoption, filepurpose="input"):
        """Checks if **fileoption** is a recursive file."""
        if self.isarchive(fileoption, filepurpose):
            return True
        return super().isrecursive(fileoption, filepurpose)

    def isarchive(self, fileoption, filepurpose="input"):
        """Returns whether the file option is an archive file."""
        if not isinstance(fileoption, str):
            return False
        mustexist = filepurpose != "output"
        if mustexist and not os.path.isfile(fileoption):
            return False
        fileext = self.splitext(fileoption)[1]
        # if None is in the archive formats, then treat all non-directory
        # inputs as archives
        return (
            self.getarchiveclass(fileext, filepurpose, os.path.isdir(fileoption))
            is not None
        )

    def getarchiveclass(self, fileext, filepurpose, isdir=False):
        """Returns the archiveclass for the given fileext and filepurpose."""
        archiveclass = self.archiveformats.get(fileext, None)
        if archiveclass is not None:
            return archiveclass
        archiveclass = self.archiveformats.get((fileext, filepurpose), None)
        if archiveclass is not None:
            return archiveclass
        if not isdir:
            archiveclass = self.archiveformats.get(None, None)
            if archiveclass is not None:
                return archiveclass
            archiveclass = self.archiveformats.get((None, filepurpose), None)
            if archiveclass is not None:
                return archiveclass
        return None

    def openarchive(self, archivefilename, filepurpose, **kwargs):
        """Creates an archive object for the given file."""
        archiveext = self.splitext(archivefilename)[1]
        archiveclass = self.getarchiveclass(
            archiveext, filepurpose, os.path.isdir(archivefilename)
        )
        archiveoptions = self.archiveoptions.copy()
        archiveoptions.update(kwargs)
        return archiveclass(archivefilename, **archiveoptions)

    def recurseinputfiles(self, options):
        """
        Recurse through archive file / directories and return files to be
        converted.
        """
        if self.isarchive(options.input, "input"):
            options.inputarchive = self.openarchive(options.input, "input")
            return self.recursearchivefiles(options)
        return super().recurseinputfiles(options)

    def recursearchivefiles(self, options):
        """Recurse through archive files and convert files."""
        inputfiles = []
        for inputpath in options.inputarchive:
            if self.isexcluded(options, inputpath):
                continue
            _top, name = os.path.split(inputpath)
            if not self.isvalidinputname(name):
                continue
            inputfiles.append(inputpath)
        return inputfiles

    def openinputfile(self, options, fullinputpath):
        """Opens the input file."""
        if self.isarchive(options.input, "input"):
            return options.inputarchive.openinputfile(fullinputpath)
        return super().openinputfile(options, fullinputpath)

    def getfullinputpath(self, options, inputpath):
        """Gets the absolute path to an input file."""
        if self.isarchive(options.input, "input"):
            return inputpath
        return os.path.join(options.input, inputpath)

    def opentemplatefile(self, options, fulltemplatepath):
        """Opens the template file (if required)."""
        if (
            fulltemplatepath is not None
            and options.recursivetemplate
            and self.isarchive(options.template, "template")
        ):
            # TODO: deal with different names in input/template archives
            if fulltemplatepath in self.templatearchive:
                return self.templatearchive.openinputfile(fulltemplatepath)
            self.warning(f"missing template file {fulltemplatepath}")
        return super().opentemplatefile(options, fulltemplatepath)

    def getfulltemplatepath(self, options, templatepath):
        """Gets the absolute path to a template file."""
        if templatepath is not None and self.usetemplates and options.template:
            if self.isarchive(options.template, "template"):
                return templatepath
            if not options.recursivetemplate:
                return templatepath
            return os.path.join(options.template, templatepath)
        return None

    def templateexists(self, options, templatepath):
        """Returns whether the given template exists..."""
        if templatepath is not None and self.isarchive(options.template, "template"):
            # TODO: deal with different names in input/template archives
            return templatepath in self.templatearchive
        return super().templateexists(options, templatepath)

    def getfulloutputpath(self, options, outputpath):
        """Gets the absolute path to an output file."""
        if self.isarchive(options.output, "output"):
            return outputpath
        if options.recursiveoutput and options.output:
            return os.path.join(options.output, outputpath)
        return outputpath

    def checkoutputsubdir(self, options, subdir) -> None:
        """
        Checks to see if subdir under ``options.output`` needs to be
        created, creates if necessary.
        """
        if not self.isarchive(options.output, "output"):
            super().checkoutputsubdir(options, subdir)

    def openoutputfile(self, options, fulloutputpath):
        """Opens the output file."""
        if self.isarchive(options.output, "output"):
            outputstream = self.outputarchive.openoutputfile(fulloutputpath)
            if outputstream is None:
                self.warning(
                    f"Could not find where to put {fulloutputpath} in output "
                    "archive; writing to tmp"
                )
                return BytesIO()
            return outputstream
        return super().openoutputfile(options, fulloutputpath)

    def recursiveprocess(self, options):
        """Recurse through directories and convert files."""
        if hasattr(options, "multifilestyle"):
            self.archiveoptions = {"multifilestyle": options.multifilestyle}
            for filetype in ("input", "output", "template"):
                allowoption = f"allowrecursive{filetype}"
                if options.multifilestyle == "onefile" and getattr(
                    options, allowoption, True
                ):
                    setattr(options, allowoption, False)

        if (
            self.usetemplates
            and options.template
            and self.isarchive(options.template, "template")
        ):
            self.templatearchive = self.openarchive(options.template, "template")

        if options.output and self.isarchive(options.output, "output"):
            self.outputarchive = self.openarchive(options.output, "output", mode="w")
        return super().recursiveprocess(options)

    def processfile(
        self, fileprocessor, options, fullinputpath, fulloutputpath, fulltemplatepath
    ):
        """Run an individual conversion."""
        if options.timestamp and _output_is_newer(fullinputpath, fulloutputpath):
            return False

        if self.isarchive(options.output, "output"):
            inputfile = self.openinputfile(options, fullinputpath)
            # TODO: handle writing back to same archive as input/template
            templatefile = self.opentemplatefile(options, fulltemplatepath)
            outputfile = self.openoutputfile(options, fulloutputpath)
            passthroughoptions = self.getpassthroughoptions(options)
            if fileprocessor(inputfile, outputfile, templatefile, **passthroughoptions):
                if not outputfile.isatty():
                    outputfile.close()
                return True
            if fulloutputpath and os.path.isfile(fulloutputpath):
                outputfile.close()
                os.unlink(fulloutputpath)
            return False
        return super().processfile(
            fileprocessor, options, fullinputpath, fulloutputpath, fulltemplatepath
        )


def _output_is_newer(input_path, output_path):
    """
    Check if input_path was not modified since output_path was generated,
    used to avoid needless regeneration of output.
    """
    if not input_path or not output_path:
        return False

    if not os.path.exists(input_path) or not os.path.exists(output_path):
        return False

    input_mtime = os.path.getmtime(input_path)
    output_mtime = os.path.getmtime(output_path)

    return output_mtime > input_mtime


def should_output_store(store, threshold):
    """
    Check if the percent of translated source words more than or equal to
    the given threshold.
    """
    if not threshold:
        return True

    units = [unit for unit in store.units if unit.istranslatable()]
    translated = [unit for unit in units if unit.istranslated()]
    wordcounts = {unit.getid(): pocount.wordsinunit(unit) for unit in units}

    def sourcewords(elementlist):
        return sum(wordcounts[unit.getid()][0] for unit in elementlist)

    translated_count = sourcewords(translated)
    total_count = sourcewords(units)
    percent = translated_count * 100 / total_count

    return percent >= threshold


class DocpathMerger:
    """
    Mixin class for merging translations using docpath-based matching.

    This mixin provides common functionality for converters that need to
    create PO files from existing translations by matching translation units
    based on their logical document path (docpath) rather than source text.
    """

    def merge_stores_by_docpath(
        self,
        inputfile,
        templatefile,
        outputstore,
        input_store_class,
        template_store_class=None,
        filter_header=False,
        process_unit_callback=None,
    ) -> None:
        """
        Merge translation from inputfile with source from templatefile using docpath matching.

        :param inputfile: File handle for the translated file (target language)
        :param templatefile: File handle for the source file (source language)
        :param outputstore: PO store to add units to
        :param input_store_class: Class to parse the input file
        :param template_store_class: Class to parse the template file (defaults to input_store_class)
        :param filter_header: Whether to skip units where isheader() returns True
        :param process_unit_callback: Optional callback to process each unit (unit, po_unit) -> None
        """
        if template_store_class is None:
            template_store_class = input_store_class

        # Parse both files - note that storage classes use 'inputfile' as parameter name
        templateparser = template_store_class(inputfile=templatefile)  # Source language
        inputparser = input_store_class(inputfile=inputfile)  # Translated language

        # Build a docpath index for the input (translated) file
        input_index = {}
        for unit in inputparser.units:
            if filter_header and hasattr(unit, "isheader") and unit.isheader():
                continue
            docpath = unit.getdocpath()
            if docpath:
                input_index[docpath] = unit

        # Iterate through template units and match with input by docpath
        for templateunit in templateparser.units:
            if (
                filter_header
                and hasattr(templateunit, "isheader")
                and templateunit.isheader()
            ):
                continue

            docpath = templateunit.getdocpath()
            storeunit = outputstore.addsourceunit(templateunit.source)
            storeunit.addlocations(templateunit.getlocations())

            # Set target from matching input unit if found
            if docpath and docpath in input_index:
                inputunit = input_index[docpath]
                storeunit.target = inputunit.source

            # Allow subclass-specific processing
            if process_unit_callback:
                process_unit_callback(templateunit, storeunit)


def main(argv=None) -> None:
    parser = ArchiveConvertOptionParser({}, description=__doc__)
    parser.run(argv)



#
# Copyright 2004-2006 Zuza Software Foundation
#
# This file is part of translate.
#
# translate is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# translate is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

"""
Convert Gettext PO localization files to Qt Linguist (.ts) files.

See: http://docs.translatehouse.org/projects/translate-toolkit/en/latest/commands/ts2po.html
for examples and usage instructions.
"""

from translate.convert import convert
from translate.misc.multistring import multistring
from translate.storage import po, ts2


class po2ts:
    @staticmethod
    def _merge_plural_forms(singular, plural):
        """
        Merge singular and plural forms into (s) notation for TS format.

        For example: "item" and "items" becomes "item(s)"
                     "day ago" and "days ago" becomes "day(s) ago"
        Returns the merged form if possible, otherwise returns the plural form.

        Note: This is necessary because the Qt Linguist TS format does not support
        separate singular and plural source forms like Gettext PO does. As a result,
        we merge the forms into a single string using (s) notation where possible.
        """
        if not singular or not plural:
            return plural or singular

        # Try to find where they differ
        min_len = min(len(singular), len(plural))

        # Find the first difference
        first_diff = min_len
        for i in range(min_len):
            if singular[i] != plural[i]:
                first_diff = i
                break

        # If they're identical up to the end of the shorter string
        if first_diff == min_len:
            # Check if plural is just singular + 's'
            if len(plural) == len(singular) + 1 and plural == f"{singular}s":
                return f"{singular}(s)"
            # Check if plural is just singular + 'es'
            if len(plural) == len(singular) + 2 and plural == f"{singular}es":
                return f"{singular}(es)"
        # They differ somewhere in the middle
        # Check if the difference is just an 's' insertion
        # e.g., "day ago" vs "days ago"
        elif (
            len(plural) == len(singular) + 1
            and singular[:first_diff] == plural[:first_diff]
            and plural[first_diff] == "s"
            and singular[first_diff:] == plural[first_diff + 1 :]
        ):
            return f"{plural[:first_diff]}(s){plural[first_diff + 1 :]}"

        # If we can't merge intelligently, return the plural form
        return plural

    @staticmethod
    def convertstore(inputstore, outputfile, templatefile=None, context=None) -> None:
        """Converts a .po file to .ts format (using a template .ts file if given)."""
        tsfile = ts2.tsfile() if templatefile is None else ts2.tsfile(templatefile)
        for inputunit in inputstore.units:
            if inputunit.isheader() or inputunit.isblank():
                continue
            source = inputunit.source
            # For plural forms, merge singular and plural into (s) notation
            if inputunit.hasplural() and isinstance(source, multistring):
                if len(source.strings) > 1:
                    singular = source.strings[0]
                    plural = source.strings[1]
                    source = po2ts._merge_plural_forms(singular, plural)
                elif len(source.strings) == 1:
                    source = source.strings[0]
                # If strings is empty, source remains as multistring (will be handled by tsunit)
            translation = inputunit.target
            comment = inputunit.getnotes("translator")
            locations = inputunit.getlocations()
            # If there are no locations, we still need to add the unit
            # Use the provided context or an empty string as default
            if not locations:
                locations = [""]
            for sourcelocation in locations:
                if context is None:
                    if sourcelocation and "#" in sourcelocation:
                        contextname = sourcelocation[: sourcelocation.find("#")]
                    elif sourcelocation:
                        contextname = sourcelocation
                    else:
                        # No location and no context provided, use empty string
                        contextname = ""
                else:
                    contextname = context
                tsunit = ts2.tsunit(source)
                tsunit.target = translation
                if not inputunit.istranslated():
                    tsunit.markfuzzy()
                elif inputunit.getnotes("developer") == "(obsolete)":
                    tsunit.set_state_n(tsunit.S_OBSOLETE)
                tsfile.addunit(tsunit, True, contextname, comment, True)
        tsfile.serialize(outputfile)


def convertpo(inputfile, outputfile, templatefile, context) -> int:
    """Reads in stdin using fromfileclass, converts using convertorclass, writes to stdout."""
    inputstore = po.pofile(inputfile)
    if inputstore.isempty():
        return 0
    convertor = po2ts()
    convertor.convertstore(inputstore, outputfile, templatefile, context)
    return 1


def main(argv=None) -> None:
    formats = {"po": ("ts", convertpo), ("po", "ts"): ("ts", convertpo)}
    parser = convert.ConvertOptionParser(
        formats, usepots=False, usetemplates=True, description=__doc__
    )
    parser.add_option(
        "-c",
        "--context",
        dest="context",
        default=None,
        help="use supplied context instead of the one in the .po file comment",
    )
    parser.passthrough.append("context")
    parser.run(argv)


if __name__ == "__main__":
    main()
